# Copyright 2021 Multiscale Modeling of Fluid Materials, TU Munich
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Schedules parameters of the integrator and solver.

The scheduler organizes the independent variables of the update equation, such
as the temperature and the step size, which are organized by multiple specific
schedulers.

"""

# Todo: Correct typing

from collections import namedtuple
from typing import Callable, Tuple

import jax.numpy as jnp
from jax import lax
from jax import random
from jax.experimental import host_callback as hcb

from jax_sgmc.util import Array

specific_scheduler = namedtuple("specific_scheduler",
                                ["init",
                                 "update",
                                 "get"])
"""Bundles the specific scheduler as described above.

Attributes:
  init_fn: Function to initialize the specific scheduler
  update_fn: Calculate the next state of the scheduler
  get_fn: Get the scheduled parameter

"""

schedule = namedtuple("schedule",
                      ["step_size",
                       "temperature",
                       "burn_in",
                       "accept"])
"""Auxiliary variables for integrator.

Attributes:
  step_size: Learning rate
  temperature: Scaling the magnitude of the additional noise
  burn_in: Bool, whether current step can be accepted
  accept: Bool, whether current sample should be saved

"""

scheduler_state = namedtuple("scheduler_state",
                             ["state",
                              "step_size_state",
                              "temperature_state",
                              "burn_in_state",
                              "thinning_state",
                              "progress_bar_state"])
"""Collects the states of the specific schedulers.

Attributes:
  state: State of the base scheduler, e.g. to keep track of current iteration
  step_size_state: State of the step size scheduler
  temperature_state: State of the temperature scheduler
  burn_in_state: State of the burn in scheduler
  thinning_state: State of thinning
  progress_bar_state: State of the progress bar
"""

static_information = namedtuple("static_information",
                                ["samples_collected"])
"""Information which is constant during the run.

Attributes:
  samples_collected: Number of samples saved.
"""

# The scheduler combines the specific scheduler. This makes it easier to
# implement only rarely used auxiliary variables by providing default values.
# The update functions are collected at a central state.

def init_scheduler(step_size: specific_scheduler = None,
                   temperature: specific_scheduler = None,
                   burn_in: specific_scheduler = None,
                   thinning: specific_scheduler = None,
                   progress_bar: bool = True,
                   progress_bar_steps: Array = 20
                   ) -> Tuple[Callable, Callable, Callable]:
  """Initializes the scheduler.

  The scheduler combines the specific schedules for each variable. It updates
  them and gets them at a central place and makes it possible to combine them or
  provide default values.

  Args:
    step_size: Triplet from step-size scheduler initialization
    temperature: Triplet from temperature scheduler initialization
    burn_in: Triplet from burn-in scheduler initialization
    thinning: Triplet from thinning scheduler initialization
    progress_bar: Show the percentage of completed steps

  Returns:
    Returns a triplet of ``(init_fn, update_fn, get_fn)``.
  """

  # Define the default values
  if step_size is None:
    step_size = polynomial_step_size(a=1, b=1, gamma=0.0)
  if temperature is None:
    temperature = constant_temperature(tau=1.0)
  if burn_in is None:
    burn_in = initial_burn_in(n=0)
  if thinning is None:
    # Accept all samples, save all samples
    thinning = specific_scheduler(
      lambda iterations: (None, iterations),
      lambda *args, **kwargs: None,
      lambda *args, **kwargs: True)

  if progress_bar:
    init_progress_bar, update_progress_bar = _progress_bar(burn_in, thinning)

  def init_fn(iterations: int,
              **scheduler_kwargs
              ) -> Tuple[scheduler_state, static_information]:

    # Initialize all the specific schedulers
    state = (0, iterations) # Start with iteration 0
    thinning_state, total_samples = thinning.init(
      iterations,
      **scheduler_kwargs.get('thinning', {}))
    burn_in_state, collected_samples = burn_in.init(
      iterations,
      **scheduler_kwargs.get('burn_in', {}))
    # If not thinning is provided, collect all samples not subject to burn in
    total_samples = min(total_samples, collected_samples)

    if progress_bar:
      pg_steps = scheduler_kwargs.get("progress_bar_steps", progress_bar_steps)
      pg_enabled = scheduler_kwargs.get("enabled", jnp.array(progress_bar))

      progress_bar_state = init_progress_bar(
        jnp.array(iterations), total_samples, pg_steps, pg_enabled)
    else:
      progress_bar_state = None

    init_state = scheduler_state(
      state=state,
      step_size_state=step_size.init(
        iterations,
        **scheduler_kwargs.get('step_size', {})),
      temperature_state=temperature.init(
        iterations,
        **scheduler_kwargs.get('temperature', {})),
      burn_in_state=burn_in_state,
      thinning_state=thinning_state,
      progress_bar_state=progress_bar_state)
    static = static_information(
      samples_collected=total_samples)
    return init_state, static

  def update_fn(state: scheduler_state, **kwargs) -> scheduler_state:
    # Keep track of current iteration
    iteration, total_iterations = state.state
    current_iteration = iteration + 1

    # Update the states
    step_size_state = step_size.update(state.step_size_state,
                                       iteration,
                                       **kwargs)
    temperature_state = temperature.update(state.temperature_state,
                                           iteration,
                                           **kwargs)
    burn_in_state = burn_in.update(state.burn_in_state,
                                   iteration,
                                   **kwargs)
    thinning_state = thinning.update(state.thinning_state,
                                     iteration,
                                     **kwargs)

    if progress_bar:
      # The burn in and thinning state are required to count the number of
      # collected samples
      progress_bar_state = update_progress_bar(
        state.progress_bar_state,
        iteration,
        burn_in_state,
        thinning_state,
        **kwargs)
    else:
      progress_bar_state = None

    new_scheduler_state = (current_iteration, total_iterations)
    updated_scheduler_state = scheduler_state(
      state=new_scheduler_state,
      step_size_state=step_size_state,
      temperature_state=temperature_state,
      burn_in_state=burn_in_state,
      thinning_state=thinning_state,
      progress_bar_state=progress_bar_state)
    return updated_scheduler_state

  def get_fn(state: scheduler_state, **kwargs) -> schedule:
    iteration, total_iterations = state.state
    current_step_size = step_size.get(state.step_size_state,
                                      iteration,
                                      **kwargs)
    current_temperature = temperature.get(state.temperature_state,
                                          iteration,
                                          **kwargs)
    current_burn_in = burn_in.get(state.burn_in_state,
                                  iteration,
                                  **kwargs)
    current_thinning = thinning.get(state.thinning_state,
                                    iteration,
                                    **kwargs)

    current_schedule = schedule(
      step_size=jnp.array(current_step_size),
      temperature=jnp.array(current_temperature),
      burn_in=jnp.array(current_burn_in),
      accept=jnp.array(current_thinning),
    )
    return current_schedule

  return init_fn, update_fn, get_fn

################################################################################
#
# Temperature
#
################################################################################

def constant_temperature(tau: Array = 1.0) -> specific_scheduler:
  """Scales the added noise with an unchanged constant.

  Args:
    tau: Scale of the added noise

  Returns:
    Returns a triplet as described above.

  """

  def init_fn(iterations: int,
              tau: Array = tau
              ) -> Array:
    del iterations
    return tau

  def update_fn(state: Array,
                iteration: int,
                **kwargs
                ) -> Array:
    del iteration, kwargs
    return state

  def get_fn(state: Array,
             iteration: int,
             **kwargs
             ) -> Array:
    del iteration, kwargs
    return state

  return specific_scheduler(init_fn, update_fn, get_fn)


def cyclic_temperature(beta: Array=1.0, k: int=1) -> specific_scheduler:
  """Cyclic switch of the temperature between 0.0 and 1.0.

  Switches temperature form 0.0 (SGD) to 1.0 (SGLD) when ratio of initial step
  size and current step size drops below beta. This scheduler is intended to
  be used with the cyclic step size scheduler.

  Args:
    beta: Ratio of current step size to initial step size when transition to SGLD
    k: Number of cycles

  Returns:
    Returns a triplet as described above
  """
  raise NotImplementedError

################################################################################
#
# Progress bar
#
################################################################################

def _progress_bar(burn_in: specific_scheduler,
                  thinning: specific_scheduler):
  """Prints the progress of the solver.

  Args:
    burn_in: Burn in scheduler to count accepted samples
    thinning: Thinning scheduler to count accepted samples

  """

  def _print_fn(info, _):
    percentage = round(
      int(info['current_iteration']) / int(info['total_iterations']) * 100)
    total_samples = int(info["total_samples"])
    collected_samples = int(info["collected_samples"])
    current_iteration = int(info["current_iteration"])
    total_iterations = int(info["total_iterations"])

    print(f"[Step {current_iteration}/{total_iterations}]"
          f"({percentage}%) Collected {collected_samples} of "
          f"{total_samples} samples...")

  def init_fn(iterations: Array,
              num_samples: Array,
              steps: Array = jnp.array(20),
              enabled: Array = jnp.array(True)
              ) -> Tuple[Array, Array, Array, Array, Array]:
    # Set already collected samples to zero
    init_state = iterations, num_samples, jnp.zeros(1), steps, enabled
    return init_state

  def step_fn(state: Tuple[Array, Array, Array, Array, Array],
              iteration: Array,
              burn_in_state,
              thinning_state,
              **kwargs
              ):
    iterations, tot_samples, collected_samples, steps, enabled = state

    # A sample is going to be saved if it is not subject to burn in and accepted
    sample_burn_in = burn_in.get(burn_in_state, iteration, **kwargs)
    sample_accepted = thinning.get(thinning_state, iteration, **kwargs)
    saved = sample_burn_in * sample_accepted
    collected_samples += saved

    info = {
      "total_iterations": iterations,
      "current_iteration": iteration,
      "total_samples": tot_samples,
      "collected_samples": collected_samples,
      "kwargs": kwargs
    }

    # Calculate number of steps until the progress should be printed out
    num_its = jnp.int_(jnp.floor(iterations / steps))

    # Return the number of collected samples as result of id_tap
    collected_samples = lax.cond(
      jnp.logical_and(jnp.mod(iteration, num_its) == 0, enabled),
      lambda arg: hcb.id_tap(_print_fn, arg, result=collected_samples),
      lambda arg: info["collected_samples"],
      info
    )

    new_state = iterations, tot_samples, collected_samples, steps, enabled
    return new_state

  return init_fn, step_fn

################################################################################
#
# Step Size
#
################################################################################

def adaptive_step_size(burn_in = 0,
                       initial_step_size = 0.05,
                       stabilization_constant = 100,
                       decay_constant = 0.75,
                       speed_constant = 0.05,
                       target_acceptance_rate=0.02):
  """Dual averaging scheme to tune step size for schemes with MH-step.

  The adaptive step size uses the dual averaging scheme to optimize the
  acceptance rate, as proposed by [1].

  [1] https://arxiv.org/abs/1111.4246

  Args:
    burn_in: Initial iterations, in which the step size should be tuned
    initial_step_size: Initial value of the step size
    speed_constant: Bigger constant stabilizes adaption against initial
      iterations
    decay_constant: Controls decay of learning rate of the step size
    speed_constant: Weights acceptance ratio statistics
    target_acceptance_rate: Desired acceptance rate

  Returns:
    Returns a specific step size scheduler.

  """

  def init(iterations: int,
           burn_in=burn_in,
           initial_step_size=initial_step_size,
           stabilization_constant=stabilization_constant,
           decay_constant=decay_constant,
           speed_constant=speed_constant,
           target_acceptance_rate=target_acceptance_rate):
    del iterations

    x_bar = jnp.log(initial_step_size)
    h_bar = 0.0
    init_state = (
      burn_in, x_bar, h_bar, target_acceptance_rate, stabilization_constant,
      decay_constant, speed_constant, jnp.log(10 * initial_step_size))
    return init_state

  def update(state: Array, iteration: int, acceptance_ratio=0.0, **kwargs):
    del kwargs
    burn_in, x_bar, h_bar, alpha, t0, kappa, gamma, mu = state
    m = iteration + 1

    h_bar *= (1 - 1/(m + t0))
    h_bar += 1/(m + t0) * (target_acceptance_rate - acceptance_ratio)

    x = mu - jnp.sqrt(m) / gamma * h_bar

    lr = jnp.power(m, -kappa)

    x_bar_old = x_bar
    x_bar *= (1 - lr)
    x_bar += lr * x

    # Only update during burn in
    x_bar = jnp.where(iteration < burn_in, x_bar, x_bar_old)

    return burn_in, x_bar, h_bar, alpha, t0, kappa, gamma, mu

  def get(state: Array, iteration: int, **kwargs):
    del iteration, kwargs
    return jnp.exp(state[1])

  return specific_scheduler(init, update, get)


def polynomial_step_size(a: Array = 1.0,
                         b: Array = 1.0,
                         gamma: Array = 0.33
                         ) -> specific_scheduler:
  """Polynomial decreasing step size schedule.

  Implements the original proposal of a polynomial step size schedule
  :math:`\epsilon = a(b + n)^{\gamma}`.

  Args:
    a: Scale of all step sizes
    b: Stabilization constant
    gamma: Decay constant

  Returns:
    Returns triplet as described above.

  """

  # The internal state is just an array, which holds the step size for all the
  # iterations

  def init_fn(iterations: int,
              a: Array = a,
              b: Array = b,
              gamma: Array = gamma
              ) -> Array:
    assert gamma >= 0, f"Gamma must be positive: gamma = {gamma}"
    assert a > 0, f"a must be positive: a = {a}"
    assert b > 0, f"b must be greater than zero: b = {b}"

    n = jnp.arange(iterations)
    unscaled = jnp.power(b + n, -gamma)
    scaled = jnp.multiply(a, unscaled)
    return scaled

  def update_fn(state: Array, iteration: int, **kwargs) -> Array:
    del iteration, kwargs
    return state

  def get_fn(state: Array, iteration: int, **kwargs) -> Array:
    del kwargs
    return state[iteration]

  return specific_scheduler(init_fn, update_fn, get_fn)


def polynomial_step_size_first_last(first: [float, Array] = 1.0,
                                    last: [float, Array] = 1.0,
                                    gamma: [float, Array] = 0.33
                                    ) -> specific_scheduler:
  """Initializes polynomial step size schedule via first and last step.

  Args:
    first: Step size in the first iteration
    last: Step size in the last iteration
    gamma: Rate of decay

  Returns:
    Returns a polynomial step size schedule defined via the first and the last
    step size.

  """

  # Calculates the required coefficients of the polynomial
  def find_ab(its, gamma, first, last):
    ginv = jnp.power(gamma, -1.0)
    fpow = jnp.power(first, -ginv) # pylint: disable=E1130
    lpow = jnp.power(last, -ginv) # pylint: disable=E1130
    apow = jnp.divide(lpow - fpow, its - 1)
    a = jnp.power(apow, -gamma)
    b = jnp.power(jnp.divide(first, a), -ginv) # pylint: disable=E1130
    return a, b

  def init_fn(iterations: int,
              first: [float, Array] = first,
              last: [float, Array] = last,
              gamma: [float, Array] = gamma
              ) -> Array:
    # Check for valid parameters
    assert gamma > 0, f"Gamma must be bigger than 0, is {gamma}"
    assert first >= last, f"The first step size must be larger than the last:" \
                         f" {first} !>= {last}"
    a, b = find_ab(iterations, gamma, first, last)
    init_fn, _, _ = polynomial_step_size(a=a, b=b, gamma=gamma)
    return init_fn(iterations)

  def update_fn(state: Array, iteration: int, **kwargs) -> Array:
    del iteration, kwargs
    return state

  def get_fn(state: Array, iteration: int, **kwargs) -> Array:
    del kwargs
    return state[iteration]

  return specific_scheduler(init_fn, update_fn, get_fn)

################################################################################
#
# Burn In
#
################################################################################

# Burn in: Return 1.0 if the sample should be accepted and 0.0 otherwise

def cyclic_burn_in(beta: Array=1.0, k:int=1):
  """Discards samples at the beginning of each cycle.

  Args:
    beta: Ratio of current and initial step size up to which burn in should be
      applied
    k: Number of cycles

  Returns:
    Returns a burn in schedule, which applies burn in to the beginning of each
    cycle.

  """
  raise NotImplementedError

def initial_burn_in(n: Array = 0) -> specific_scheduler:
  """Discards the first n steps.

  Args:
    n: Count of initial steps which should be discarded

  Returns:
    Returns specific scheduler.

  """

  def init_fn(iterations: int, n: Array = n) -> Tuple[Array, Array]:
    return n, iterations - n

  def update_fn(state: Array, iteration: int, **kwargs) -> Array:
    del iteration, kwargs
    return state

  def get_fn(state: Array, iteration: int, **kwargs) -> Array:
    del kwargs
    return jnp.where(state <= iteration, 1.0, 0.0)

  return specific_scheduler(init_fn, update_fn, get_fn)

################################################################################
#
# Thinning
#
################################################################################

# Thinning provides information about the number of samples which will be saved.

def random_thinning(step_size_schedule: specific_scheduler,
                    burn_in_schedule: specific_scheduler,
                    selections: int,
                    key: Array = None
                    ) -> specific_scheduler:
  """Random thinning weighted by the step size.

  Randomly select samples not subject to burn in. The probability of selection
  is proportional to the step size to deal with the issue of the decaying step
  size. This only works for static step size and burn in schedules.

  Args:
    step_size_schedule: Static step size schedule
    burn_in_schedule: Static burn in schedule
    selections: Number of selected samples
    key: PRNGKey for drawing selections

  Returns:
    Returns a scheduler marking the accepted samples.

  """

  def init_fn(iterations: int,
              step_size_schedule: specific_scheduler = step_size_schedule,
              burn_in_schedule: specific_scheduler = burn_in_schedule,
              selections: int = selections,
              key: Array = key
              ) -> Tuple[Array, Array]:
    if key is None:
      key = random.PRNGKey(0)

    step_size_state = step_size_schedule.init(iterations)
    burn_in_state, _ = burn_in_schedule.init(iterations)

    def update_fn(state, iteration):
      step_size_state, burn_in_state = state
      step_size = step_size_schedule.get(step_size_state, iteration)
      burn_in = burn_in_schedule.get(burn_in_state, iteration)
      probability = step_size * burn_in
      new_state = (step_size_schedule.update(step_size_state, iteration),
                   burn_in_schedule.update(burn_in_state, iteration))
      return new_state, probability

    _, probs = lax.scan(update_fn,
                        (step_size_state, burn_in_state),
                        jnp.arange(iterations))

    # Check that a sufficient number of elements can be drawn
    assert jnp.count_nonzero(probs) >= selections, "Cannot select enough values"

    # Draw the iterations which should be accepted
    accepted_its = random.choice(key,
                                 jnp.arange(iterations),
                                 shape=(selections,),
                                 replace=False,
                                 p=probs)
    return accepted_its, selections

  def update_fn(state: Array, iteration: int, **kwargs) -> Array:
    del iteration, kwargs
    return state

  def get_fn(state: Array, iteration: int, **kwargs) -> jnp.bool_:
    del kwargs
    accepted = jnp.where(jnp.any(iteration == state), True, False)
    return accepted

  return specific_scheduler(init_fn, update_fn, get_fn)



from functools import partial


import jax
import jax.numpy as jnp
from jax import random
from jax import flatten_util
from jax import tree_util

from jax_sgmc import util
from jax_sgmc import data
from jax_sgmc.util import testing

# Todo: Test vmap on custom host_callback

import pytest



@pytest.fixture
def random_tree():
  key = random.PRNGKey(0)
  split1, split2 = random.split(key)
  tree = {"a": random.normal(split1, shape=(2,)),
          "b": {"b1": 0.0,
                "b2": random.normal(split2, shape=(3, 4))}}

  flat_tree, unravel_fn = flatten_util.ravel_pytree(tree)
  ravel_fn = lambda t: flatten_util.ravel_pytree(t)[0]
  return (tree, flat_tree), (ravel_fn, unravel_fn)


class TestTree:

  def test_tree_scale(self, random_tree):
    (tree, flat_tree), (ravel_fn, unravel_fn) = random_tree
    alpha = random.normal(random.PRNGKey(1))

    true_result = unravel_fn(alpha * flat_tree)
    treemap_result = util.tree_scale(alpha, tree)

    testing.assert_equal(true_result, treemap_result)

  def test_tree_multiply(self, random_tree):
    (tree, flat_tree), (ravel_fn, unravel_fn) = random_tree
    rnd = random.normal(random.PRNGKey(1), shape=flat_tree.shape)

    true_result = unravel_fn(jnp.multiply(rnd, flat_tree))
    treemap_result = util.tree_multiply(unravel_fn(rnd), tree)

    testing.assert_equal(true_result, treemap_result)

  def test_tree_add(self, random_tree):
    (tree, flat_tree), (ravel_fn, unravel_fn) = random_tree
    rnd = random.normal(random.PRNGKey(1), shape=flat_tree.shape)

    true_result = unravel_fn(jnp.add(rnd, flat_tree))
    treemap_result = util.tree_add(unravel_fn(rnd), tree)

    testing.assert_equal(true_result, treemap_result)

  def test_tree_matmul(self, random_tree):
    (tree, flat_tree), (ravel_fn, unravel_fn) = random_tree
    rnd = random.normal(random.PRNGKey(1), shape=(flat_tree.size, flat_tree.size))

    true_result = unravel_fn(jnp.matmul(rnd, flat_tree))
    treemap_result = util.tree_matmul(rnd, tree)

    testing.assert_equal(true_result, treemap_result)

  def test_tree_dot(self, random_tree):
    (tree, flat_tree), (ravel_fn, unravel_fn) = random_tree

    true_result = jnp.dot(flat_tree, flat_tree)
    treemap_result = util.tree_dot(tree, tree)

    testing.assert_equal(true_result, treemap_result)

class TestTreeMap():

  def test_vmap(self, random_tree):
    (random_tree, _), _ = random_tree

    zero_tree = tree_util.tree_map(jnp.zeros_like, random_tree)
    one_tree = tree_util.tree_map(jnp.ones_like, random_tree)
    modified_tree = tree_util.tree_map(jnp.add, random_tree, one_tree)

    @jax.jit
    @util.list_vmap
    def test_substract(tree):
      return tree_util.tree_map(jnp.subtract, modified_tree, tree)

    one, zero = test_substract(random_tree, modified_tree)

    testing.assert_close(zero, zero_tree)
    testing.assert_close(one, one_tree)

  @pytest.mark.pmap
  def test_pmap(self, random_tree):
    (random_tree, _), _ = random_tree

    zero_tree = tree_util.tree_map(jnp.zeros_like, random_tree)
    one_tree = tree_util.tree_map(jnp.ones_like, random_tree)
    modified_tree = tree_util.tree_map(jnp.add, random_tree, one_tree)

    @jax.jit
    @util.list_pmap
    def test_substract(tree):
      return tree_util.tree_map(jnp.subtract, modified_tree, tree)

    one, zero = test_substract(random_tree, modified_tree)

    testing.assert_close(zero, zero_tree)
    testing.assert_close(one, one_tree)


# Copyright 2021 Multiscale Modeling of Fluid Materials, TU Munich
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Popular solvers ready to use.

While JaxSGMC has been designed to be flexible, starting with full
flexibility can be complicated. Therefore, this file contains some popular
solvers with preset properties, which can be applied directly to the problem or
used as a guide to set up a custom solver.
"""
import warnings
from functools import partial
from typing import Any, Union

from jax_sgmc import data, potential, adaption, integrator, scheduler, solver, io

Pytree = Any

def sgld(potential_fn: potential.minibatch_potential,
         data_loader: data.DataLoader,
         cache_size: int = 512,
         batch_size: int = 32,
         first_step_size: float = 0.05,
         last_step_size: float = 0.001,
         burn_in: int = 0,
         accepted_samples: int = 1000,
         rms_prop: bool = False,
         alpha: float = 0.9,
         lmbd: float = 1e-5,
         save_to_numpy: bool = True,
         progress_bar: bool = True):
  """Stochastic Gradient Langevin Dynamics.

  SGLD with a polynomial step size schedule and optional speed up via RMS-prop
  adaption [1].

  [1] https://arxiv.org/abs/1512.07666

  ::

    rms_run = alias.sgld(...)

    sample = {"w": jnp.zeros((N, 1)), "sigma": jnp.array(10.0)}
    results = rms_run(sample, init_model_state=0, iterations=50000)[0]['samples']['variables']

  Args:
    potential_fn: Stochastic potential over a minibatch of data
    data_loader: Data loader, e. g. numpy data loader
    cache_size: Number of mini_batches in device memory
    batch_size: Number of observations per batch
    first_step_size: First step size
    last_step_size: Final step size
    burn_in: Number of samples to skip before collecting samples
    accepted_samples: Total number of samples to collect, will be determined by
      random thinning if accepted samples < iterations - burn_in
    rms_prop: Whether to adapt a manifold via the RMSprop strategy
    alpha: Decay speed of previous manifold approximations
    lmbd: Stabilization parameter
    save_to_numpy: Save on host in numpy array instead of in device memory
    progress_bar: Print the progress of the solver

  Returns:
    Returns a solver function which can be applied to multiple chains starting
    at ``init_sample``. If the likelihood is stateful, an initial state must be
    provided.

  """

  random_data = data.random_reference_data(data_loader, cache_size, batch_size)

  if rms_prop:
    rms_prop = adaption.rms_prop()
  else:
    rms_prop = None

  rms_integrator = integrator.langevin_diffusion(
    potential_fn, random_data, adaption=rms_prop)

  step_size_schedule = scheduler.polynomial_step_size_first_last(
    first=first_step_size, last=last_step_size)
  burn_in_schedule = scheduler.initial_burn_in(burn_in)
  random_thinning_schedule = scheduler.random_thinning(
    step_size_schedule, burn_in_schedule, selections=accepted_samples)
  schedule = scheduler.init_scheduler(
    step_size=step_size_schedule,
    burn_in=burn_in_schedule,
    thinning=random_thinning_schedule,
    progress_bar=progress_bar)

  if save_to_numpy:
    data_collector = io.MemoryCollector()
    saving = io.save(data_collector)
  else:
    saving = None

  sgld_solver = solver.sgmc(rms_integrator)
  mcmc = solver.mcmc(sgld_solver, schedule, strategy='map', saving=saving)

  def run_fn(*init_samples, init_model_state: Pytree = None, iterations = 1000):
    init_with_adaption_kwargs = partial(
      sgld_solver[0],
      adaption_kwargs={
        'alpha': alpha,
        'lmbd': lmbd,
      },
      init_model_state=init_model_state)
    states = map(init_with_adaption_kwargs, init_samples)
    return mcmc(*states, iterations=iterations)
  return run_fn

def re_sgld(potential_fn: potential.minibatch_potential,
            data_loader: data.DataLoader,
            cache_size: int = 512,
            batch_size: int = 32,
            temperature: float = 1000.0,
            first_step_size: float = 0.05,
            last_step_size: float = 0.001,
            burn_in: int = 0,
            accepted_samples: int = 100,
            save_to_numpy: bool = True,
            progress_bar: bool = True):
  """Replica Exchange Stochastic Gradient Langevin Diffusion.

  reSGLD simulates a tempered and a default chain in parallel, which
  exchange samples at random following a (biased) markov jump process [1].

  [1] https://arxiv.org/abs/2008.05367v3

  ::

    resgld_run = alias.re_sgld(...)

    sample = {"w": jnp.zeros((N, 1)), "sigma": jnp.array(2.0)}
    init_samples = [(sample, sample), (sample, sample), (sample, sample)]

    results = resgld_run(
      *init_samples, init_model_state=0, iterations=50000
      )[0]['samples']['variables']

  Args:
    potential_fn: Stochastic potential over a minibatch of data
    data_loader: Data loader, e. g. numpy data loader
    cache_size: Number of mini_batches in device memory
    batch_size: Number of observations per batch
    temperature: Temperature at which the helper chain should run
    first_step_size: First step size
    last_step_size: Final step size
    burn_in: Number of samples to skip before collecting samples
    accepted_samples: Total number of samples to collect, will be determined by
      random thinning if accepted samples < iterations - burn_in
    save_to_numpy: Save on host in numpy array instead of in device memory
    progress_bar: Print the progress of the solver


  Returns:
    Returns a solver function which can be applied to multiple chains starting
    at ``start_chain_{idx}``.

  """
  del progress_bar

  random_data = data.random_reference_data(data_loader, cache_size, batch_size)

  resgld_integrator = integrator.langevin_diffusion(
    potential_fn, random_data)

  step_size_schedule = scheduler.polynomial_step_size_first_last(
    first=first_step_size, last=last_step_size)
  burn_in_schedule = scheduler.initial_burn_in(burn_in)
  random_thinning_schedule = scheduler.random_thinning(
    step_size_schedule, burn_in_schedule, selections=accepted_samples)
  temperature_schedule = scheduler.constant_temperature(1.0)
  schedule = scheduler.init_scheduler(
    step_size=step_size_schedule,
    burn_in=burn_in_schedule,
    thinning=random_thinning_schedule,
    temperature=temperature_schedule,
    progress_bar=False)

  if save_to_numpy:
    data_collector = io.MemoryCollector()
    saving = io.save(data_collector)
  else:
    saving = None

  resgld_solver = solver.parallel_tempering(resgld_integrator)
  mcmc = solver.mcmc(resgld_solver, schedule, strategy='map', saving=saving)

  def run_fn(*init_samples, init_model_state: Pytree = None, iterations = 1000):
    init_resgld_fn = partial(
      resgld_solver[0],
      init_model_state=init_model_state)
    states = map(init_resgld_fn, *zip(*init_samples))
    return mcmc(*states,
                iterations=iterations,
                schedulers=[{'temperature': {'tau': temperature}}])
  return run_fn

def amagold(stochastic_potential_fn: potential.StochasticPotential,
            full_potential_fn: potential.FullPotential,
            data_loader: data.DataLoader,
            cache_size: int = 512,
            batch_size: int = 32,
            integration_steps: int = 10,
            friction: float = 0.25,
            first_step_size: float = 0.001,
            last_step_size: float = 0.001,
            adaptive_step_size: bool = False,
            stabilization_constant: int = 10,
            decay_constant: float = 0.75,
            speed_constant: float = 0.05,
            target_acceptance_rate: float = 0.25,
            burn_in: int = 0,
            accepted_samples: Union[int, None] = None,
            mass: Pytree = None,
            save_to_numpy: bool = True,
            progress_bar: bool = True):
  """Amortized Metropolis Adjustment for Efficient Stochastic Gradient MCMC.

  The AMAGOLD solver constructs a skew-reversible markov chain, such that
  MH-correction steps can be applied periodically to sample from the correct
  distribution [1].

  [1] https://arxiv.org/abs/2003.00193

  ::

    amagold_run = alias.amagold(...)

    sample = {"w": jnp.zeros((N, 1)), "sigma": jnp.array(2.0)}

    results = amagold_run(
      sample, init_model_state=0, iterations=5000
      )[0]['samples']['variables']

  Args:
    stochastic_potential_fn: Stochastic potential over a minibatch of data
    full_potential_fn: Potential from full dataset
    data_loader: Data loader, e. g. numpy data loader
    cache_size: Number of mini_batches in device memory
    batch_size: Number of observations per batch
    integration_steps: Number of leapfrog-steps before each MH-correction step
    friction: Parameter between 0.0 and 1.0 controlling decay of momentum
    first_step_size: First step size for polynomial and adaptive step size
      schedule
    last_step_size: Final step size of the polynomial step size schedule
    adaptive_step_size: Adapt the step size to optimize the acceptance rate
      during burn in
    stabilization_constant: Larger numbers reduce the impact of the initial
      steps on the step size
    decay_constant: Larger values reduce impact of later steps
    speed_constant: Speed of adaption of the step size
    target_acceptance_rate: Target of the adption of the step sizes
    burn_in: Number of samples to skip before collecting samples
    accepted_samples: Total number of samples to collect, will be determined by
      random thinning if accepted samples < iterations - burn_in.
      If None, no thinning wil be applied.
    mass: Diagonal mass for HMC-dynamics
    save_to_numpy: Save on host in numpy array instead of in device memory
    progress_bar: Print the progress of the solver

  Returns:
    Returns a solver function which can be applied to multiple chains starting
    at ``init_sample``.

  """

  random_data = data.random_reference_data(data_loader, cache_size, batch_size)
  full_data_map = data.full_reference_data(data_loader, cache_size, batch_size)

  reversible_leapfrog = integrator.reversible_leapfrog(
    stochastic_potential_fn, random_data, integration_steps, friction, mass)
  amagold_solver = solver.amagold(
    reversible_leapfrog, full_potential_fn, full_data_map)

  burn_in_schedule = scheduler.initial_burn_in(burn_in)
  if adaptive_step_size:
    step_size_schedule = scheduler.adaptive_step_size(
      burn_in=burn_in,
      initial_step_size=first_step_size,
      stabilization_constant=stabilization_constant,
      decay_constant=decay_constant,
      speed_constant=speed_constant,
      target_acceptance_rate=target_acceptance_rate)
    random_thinning_schedule = None
    assert accepted_samples is None, ('Thinning currently not supported for'
                                      ' adaptive step size.')
  else:
    step_size_schedule = scheduler.polynomial_step_size_first_last(
      first=first_step_size,
      last=last_step_size)
    if accepted_samples is None:
      random_thinning_schedule = None
    else:
      random_thinning_schedule = scheduler.random_thinning(
        step_size_schedule, burn_in_schedule, selections=accepted_samples)
  schedule = scheduler.init_scheduler(
    step_size=step_size_schedule,
    burn_in=burn_in_schedule,
    thinning=random_thinning_schedule,
    progress_bar=progress_bar)

  if save_to_numpy:
    data_collector = io.MemoryCollector()
    saving = io.save(data_collector)
  else:
    saving = None

  mcmc = solver.mcmc(amagold_solver, schedule, strategy='map', saving=saving)

  def run_fn(*init_samples, init_model_state: Pytree = None, iterations=1000):
    init_amagold_fn = partial(
      amagold_solver[0],
      init_model_state = init_model_state)
    states = map(init_amagold_fn, init_samples)
    return mcmc(*states, iterations=iterations)
  return run_fn

def sggmc(stochastic_potential_fn: potential.StochasticPotential,
          full_potential_fn: potential.FullPotential,
          data_loader: data.DataLoader,
          cache_size: int = 512,
          batch_size: int = 32,
          integration_steps: int = 10,
          friction_coefficient: float = 1.0,
          first_step_size: float = 0.001,
          last_step_size: float = 0.001,
          adaptive_step_size: bool = False,
          stabilization_constant: int = 10,
          decay_constant: float = 0.75,
          speed_constant: float = 0.05,
          target_acceptance_rate: float = 0.25,
          burn_in: int = 0,
          accepted_samples: Union[int, None] = None,
          mass: Pytree = None,
          save_to_numpy: bool = True,
          progress_bar: bool = True):
  """Stochastic gradient guided monte carlo.

  The SGGMC solver is based on the OBABO integrator, which is reversible
  when using stochastic gradients. Moreover, the calculation of the full
  potential is only necessary once per MH-correction step, which can be applied
  after multiple iterations [1].

  [1] https://arxiv.org/abs/2102.01691

  ::

    sggmc_run = alias.sggmc(...)

    sample = {"w": jnp.zeros((N, 1)), "sigma": jnp.array(2.0)}

    results = sggmc_run(
      sample, init_model_state=0, iterations=5000
      )[0]['samples']['variables']

  Args:
    stochastic_potential_fn: Stochastic potential over a minibatch of data
    full_potential_fn: Potential from full dataset
    data_loader: Data loader, e. g. numpy data loader
    cache_size: Number of mini_batches in device memory
    batch_size: Number of observations per batch
    integration_steps: Number of leapfrog-steps before each MH-correction step
    friction_coefficient: Positive parameter controling amount of refreshed
     momentum
    first_step_size: First step size for polynomial and adaptive step size
      schedule
    last_step_size: Final step size of the polynomial step size schedule
    adaptive_step_size: Adapt the step size to optimize the acceptance rate
      during burn in
    stabilization_constant: Larger numbers reduce the impact of the initial
      steps on the step size
    decay_constant: Larger values reduce impact of later steps
    speed_constant: Speed of adaption of the step size
    target_acceptance_rate: Target of the adption of the step sizes
    burn_in: Number of samples to skip before collecting samples
    accepted_samples: Total number of samples to collect, will be determined by
      random thinning if accepted samples < iterations - burn_in.
      If None, no thinning wil be applied.
    mass: Diagonal mass for HMC-dynamics
    save_to_numpy: Save on host in numpy array instead of in device memory
    progress_bar: Print the progress of the solver

  Returns:
    Returns a solver function which can be applied to multiple chains starting
    at ``init_sample``.

  """

  random_data = data.random_reference_data(data_loader, cache_size, batch_size)
  full_data_map = data.full_reference_data(data_loader, cache_size, batch_size)

  obabo = integrator.obabo(
    stochastic_potential_fn, random_data, integration_steps,
    friction_coefficient, mass)
  sggmc_solver = solver.sggmc(
    obabo, full_potential_fn, full_data_map)

  burn_in_schedule = scheduler.initial_burn_in(burn_in)
  if adaptive_step_size:
    step_size_schedule = scheduler.adaptive_step_size(
      burn_in=burn_in,
      initial_step_size=first_step_size,
      stabilization_constant=stabilization_constant,
      decay_constant=decay_constant,
      speed_constant=speed_constant,
      target_acceptance_rate=target_acceptance_rate)
    random_thinning_schedule = None
    assert accepted_samples is None, ('Thinning currently not supported for'
                                      ' adaptive step size.')
  else:
    step_size_schedule = scheduler.polynomial_step_size_first_last(
      first=first_step_size,
      last=last_step_size)
    if accepted_samples is None:
      random_thinning_schedule = None
    else:
      random_thinning_schedule = scheduler.random_thinning(
        step_size_schedule, burn_in_schedule, selections=accepted_samples)
  schedule = scheduler.init_scheduler(
    step_size=step_size_schedule,
    burn_in=burn_in_schedule,
    thinning=random_thinning_schedule,
    progress_bar=progress_bar)

  if save_to_numpy:
    data_collector = io.MemoryCollector()
    saving = io.save(data_collector)
  else:
    saving = None

  mcmc = solver.mcmc(sggmc_solver, schedule, strategy='map', saving=saving)

  def run_fn(*init_samples, init_model_state: Pytree = None, iterations=1000):
    init_sggmc_fn = partial(sggmc_solver[0], init_model_state=init_model_state)
    states = map(init_sggmc_fn, init_samples)
    return mcmc(*states, iterations=iterations)
  return run_fn

def sghmc(potential_fn: potential.minibatch_potential,
          data_loader: data.DataLoader,
          cache_size: int = 512,
          batch_size: int = 32,
          integration_steps: int = 10,
          friction: Union[float, Pytree] = 1.0,
          mass: Pytree = None,
          first_step_size: float = 0.05,
          last_step_size: float = 0.001,
          burn_in: int = 0,
          accepted_samples: int = 1000,
          adapt_noise_model: bool = False,
          diagonal_noise: bool = True,
          save_to_numpy: bool = True,
          progress_bar: bool = True):
  """Stochastic Gradient Hamiltonian Monte Carlo.

  SGHMC improves the exploratory power of SGLD by introducing momentum [1].

  [1] https://arxiv.org/abs/1402.4102

  ::

    sghmc_run = alias.sghmc(...)

    sample = {"w": jnp.zeros((N, 1)), "sigma": jnp.array(2.0)}
    results = sghmc_run(sample, init_model_state=0, iterations=5000)[0]['samples']['variables']

  Args:
    potential_fn: Stochastic potential over a minibatch of data
    data_loader: Data loader, e. g. numpy data loader
    cache_size: Number of mini_batches in device memory
    batch_size: Number of observations per batch
    integration_steps: Number of leapfrog steps before resampling the momentum
    friction: Friction to counteract noise introduced by stochastic gradients.
      Can be specified for each variable or for all variables (scalar value)
    mass: Diagonal mass to be used for hamiltonian dynamics
    first_step_size: First step size
    last_step_size: Final step size
    burn_in: Number of samples to skip before collecting samples
    accepted_samples: Total number of samples to collect, will be determined by
      random thinning if accepted samples < iterations - burn_in
    adapt_noise_model: Estimate the gradient noise to speed up the convergence.
    diagonal_noise: Restrict the noise estimate to be diagonal.
    save_to_numpy: Save on host in numpy array instead of in device memory
    progress_bar: Print the progress of the solver

  Returns:
    Returns a solver function which can be applied to multiple chains starting
    at ``init_sample``.

  """

  random_data = data.random_reference_data(data_loader, cache_size, batch_size)

  if adapt_noise_model:
    noise_model = adaption.fisher_information(minibatch_potential=potential_fn,
                                              diagonal=diagonal_noise)
  else:
    noise_model = None

  friction_leapfrog = integrator.friction_leapfrog(
    potential_fn, random_data, friction=friction, const_mass=mass,
    steps=integration_steps, noise_model=noise_model)

  step_size_schedule = scheduler.polynomial_step_size_first_last(
    first=first_step_size, last=last_step_size)
  burn_in_schedule = scheduler.initial_burn_in(burn_in)
  random_thinning_schedule = scheduler.random_thinning(
    step_size_schedule, burn_in_schedule, selections=accepted_samples)
  schedule = scheduler.init_scheduler(
    step_size=step_size_schedule,
    burn_in=burn_in_schedule,
    thinning=random_thinning_schedule,
    progress_bar=progress_bar)

  if save_to_numpy:
    data_collector = io.MemoryCollector()
    saving = io.save(data_collector)
  else:
    saving = None

  sghmc_solver = solver.sgmc(friction_leapfrog)
  mcmc = solver.mcmc(sghmc_solver, schedule, strategy='map', saving=saving)

  def run_fn(*init_samples, init_model_state: Pytree = None, iterations = 1000):
    init_sghmc_fn = partial(sghmc_solver[0], init_model_state=init_model_state)
    states = map(init_sghmc_fn, init_samples)
    return mcmc(*states, iterations=iterations)
  return run_fn

def obabo(potential_fn: potential.minibatch_potential,
          data_loader: data.DataLoader,
          cache_size: int = 512,
          batch_size: int = 32,
          integration_steps: int = 10,
          friction: Union[float, Pytree] = 1.0,
          mass: Pytree = None,
          first_step_size: float = 0.05,
          last_step_size: float = 0.001,
          burn_in: int = 0,
          accepted_samples: int = 1000,
          save_to_numpy: bool = True,
          progress_bar: bool = True):
  """Langevin Monte Carlo with partial momentum refreshment.

  [1] https://arxiv.org/abs/2102.01691

  ::

    sghmc_run = alias.obabo(...)

    sample = {"w": jnp.zeros((N, 1)), "sigma": jnp.array(2.0)}
    results = sghmc_run(sample, init_model_state=0, iterations=5000)[0]['samples']['variables']

  Args:
    potential_fn: Stochastic potential over a minibatch of data
    data_loader: Data loader, e. g. numpy data loader
    cache_size: Number of mini_batches in device memory
    batch_size: Number of observations per batch
    integration_steps: Number of leapfrog steps before resampling the momentum
    friction: Positive parameter controling amount of refreshed momentum
    mass: Diagonal mass to be used for hamiltonian dynamics
    first_step_size: First step size
    last_step_size: Final step size
    burn_in: Number of samples to skip before collecting samples
    accepted_samples: Total number of samples to collect, will be determined by
      random thinning if accepted samples < iterations - burn_in
    save_to_numpy: Save on host in numpy array instead of in device memory
    progress_bar: Print the progress of the solver

  Returns:
    Returns a solver function which can be applied to multiple chains starting
    at ``init_sample``.

  """

  random_data = data.random_reference_data(data_loader, cache_size, batch_size)

  obabo_integrator = integrator.obabo(
    potential_fn=potential_fn, batch_fn=random_data, steps=integration_steps,
    friction=friction, const_mass=mass
  )

  step_size_schedule = scheduler.polynomial_step_size_first_last(
    first=first_step_size, last=last_step_size)
  burn_in_schedule = scheduler.initial_burn_in(burn_in)
  random_thinning_schedule = scheduler.random_thinning(
    step_size_schedule, burn_in_schedule, selections=accepted_samples)
  schedule = scheduler.init_scheduler(
    step_size=step_size_schedule,
    burn_in=burn_in_schedule,
    thinning=random_thinning_schedule,
    progress_bar=progress_bar)

  if save_to_numpy:
    data_collector = io.MemoryCollector()
    saving = io.save(data_collector)
  else:
    saving = None

  sghmc_solver = solver.sgmc(obabo_integrator)
  mcmc = solver.mcmc(sghmc_solver, schedule, strategy='map', saving=saving)

  def run_fn(*init_samples, init_model_state: Pytree = None, iterations = 1000):
    init_sghmc_fn = partial(sghmc_solver[0], init_model_state=init_model_state)
    states = map(init_sghmc_fn, init_samples)
    return mcmc(*states, iterations=iterations)
  return run_fn



"""Test io module. """
from functools import partial

from jax import random, jit, pmap, vmap, lax, tree_map
import jax.numpy as jnp

import pytest

from jax_sgmc import io, scheduler, util
from jax_sgmc.util import testing

class TestDictTransformation:
  pass







@pytest.mark.hdf5
class TestHDF5Collector:
  pass






class TestMemoryCollector:
  pass








class TestSaving:
  """Test saving."""

  @pytest.fixture
  def test_function(self):

    shape = (4, 3)

    def _update(state, _):
      key, split1, split2 = random.split(state["key"], 3)
      noise = random.normal(split1, shape=shape)
      new_state = {"key": key,
                   "results": {"noise": noise,
                               "sum": state["results"]["sum"] + noise}}
      return new_state, new_state

    def _init(seed):
      init_state = {"key": random.PRNGKey(seed),
                    "results": {"noise": jnp.zeros(shape),
                                "sum": jnp.zeros(shape)}}
      return init_state

    def _run(init_state, keep):
      _, all_results = lax.scan(_update, init_state, keep)
      accepted_results = tree_map(lambda leaf: leaf[keep], all_results)
      return accepted_results

    return _init, _update, _run

  def test_no_save(self, test_function):
    """Test saving by running save against no_save and direct output on random
    operations.
    """

    # Calculate the true results
    _init, _update, _run = test_function
    accepted_list = random.bernoulli(random.PRNGKey(11), shape=(15,))
    count = jnp.sum(accepted_list)

    init_state = _init(0)
    reference_solution = _run(init_state, accepted_list)

    # Run the no_save solution

    init_save, save, postprocess_save = io.no_save()

    def no_save_run(init_sample):
      def update(state, keep):
        saving_state, simulation_state = state
        simulation_state, sample = _update(simulation_state, None)
        saving_state, _ = save(saving_state, keep, simulation_state)
        return (saving_state, simulation_state), None
      saving_state = init_save(init_sample, {}, io.scheduler.static_information(samples_collected=count))
      (saving_state, _), _ = lax.scan(update, (saving_state, init_sample), accepted_list)
      return postprocess_save(saving_state, None)["samples"]

    no_save_results = no_save_run(init_state)

    # Check close

    testing.assert_equal(no_save_results, reference_solution)

  def test_save(self, test_function):
    """Test saving by running save against no_save and direct output on random
    operations.
    """

    # Calculate the true results
    _init, _update, _run = test_function
    accepted_list = random.bernoulli(random.PRNGKey(11), shape=(15,))
    count = jnp.sum(accepted_list)

    init_state = _init(0)
    reference_solution = _run(init_state, accepted_list)

    # Run the no_save solution

    data_collector = io.MemoryCollector()
    init_save, save, postprocess_save = io.save(data_collector)

    def save_run(init_sample):
      def update(state, keep):
        saving_state, simulation_state = state
        simulation_state, sample = _update(simulation_state, None)
        saving_state, _ = save(saving_state, keep, simulation_state)
        return (saving_state, simulation_state), None
      saving_state = init_save(init_sample, {}, io.scheduler.static_information(samples_collected=count))
      (saving_state, _), _ = lax.scan(update, (saving_state, init_sample), accepted_list)
      return postprocess_save(saving_state, None)["samples"]

    save_results = save_run(init_state)

    # Check close

    testing.assert_equal(save_results, reference_solution)

  @pytest.mark.skip
  def test_save_vmap(self, test_function):
    """Test saving by running save against no_save and direct output on random
    operations.
    """

    # Calculate the true results
    _init, _update, _run = test_function
    accepted_list = random.bernoulli(random.PRNGKey(11), shape=(15,))
    count = jnp.sum(accepted_list)
    seeds = jnp.arange(2)


    init_states = list(map(_init, seeds))
    reference_solution = list(map(lambda state: _run(state, accepted_list), init_states))

    # Run the no_save solution

    data_collector = io.MemoryCollector()
    init_save, save, postprocess_save = io.save(data_collector)

    vmap_init_states = [((init_save(init_sample, {}, io.scheduler.static_information(samples_collected=count)),
                         init_sample), accepted_list) for init_sample in init_states]

    # We use list_vmap, because we tun the chain over a list of pytrees. Also,
    # we currently have to vectorize over the acceptance list because of the
    # implementation of stop_vmap
    @util.list_vmap
    def save_run(init_state):
      init_args, accept = init_state
      saving_state, init_sample = init_args
      def update(state, keep):
        saving_state, simulation_state = state
        simulation_state, sample = _update(simulation_state, None)
        saving_state, _ = save(saving_state, keep, simulation_state)
        return (saving_state, simulation_state), None

      (saving_state, _), _ = lax.scan(update, (saving_state, init_sample), accept)
      return saving_state

    save_results = [postprocess_save(res, None)["samples"] for res in save_run(*vmap_init_states)]

    # Check close

    testing.assert_equal(save_results, reference_solution)



# Copyright 2021 Multiscale Modeling of Fluid Materials, TU Munich
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Use samples saved with :class:`jax_sgmc.io.HDF5Collector` as reference data. """

import itertools
from typing import Any

import h5py
import numpy as onp
import jax.numpy as jnp
import jax

from jax_sgmc.data.numpy_loader import NumpyDataLoader
from jax_sgmc.io import pytree_dict_keys

PyTree = Any

# Inherit from NumpyDataLoader because slicing of arrays is similar
class HDF5Loader(NumpyDataLoader):
  """Load reference data from HDF5-files.

  This data loader can load reference data stored in HDF5 files. This makes it
  possible to use the :mod:`jax_sgmc.data` module to evaluate samples saved via
  the :class:`jax_sgmc.io.HDF5Collector`.

  Args:
    file: Path to the HDF5 file containing the reference data
    subdir: Path to the subset of the data set which should be loaded
    sample: PyTree to specify the original shape of the sub-pytree before it
      has been saved by the :class:`jax_sgmc.io.HDF5Collector`

   """

  def __init__(self, file, subdir="/chain~0/variables/", sample=None):
    # The sample is necessary to return the observations in the correct format.
    super().__init__()

    if isinstance(file, h5py.File):
      self._dataset = file
    else:
      self._dataset = h5py.File(name=file, mode="r")
    self._reference_data = ["/".join(itertools.chain([subdir], key_tuple))
                            for key_tuple in pytree_dict_keys(sample)]
    self._pytree_structure = jax.tree_structure(sample)
    self._sample_format = jax.tree_map(
      lambda leaf: jax.ShapeDtypeStruct(shape=leaf.shape, dtype=leaf.dtype),
      sample)

    observations_counts = [len(self._dataset[leaf_name])
                           for leaf_name in self._reference_data]
    self._observation_count = observations_counts[0]

  def get_batches(self, chain_id: int) -> PyTree:
    """Draws a batch from a chain.

    Args:
      chain_id: ID of the chain, which holds the information about the form of
        the batch and the process of assembling.

    Returns:
      Returns a superbatch as registered by :func:`register_random_pipeline` or
      :func:`register_ordered_pipeline` with `cache_size` batches holding
      `mb_size` observations.

    """
    # Data slicing is the same for all methods of random and ordered access,
    # only the indices for slicing differ. The method _get_indices find the
    # correct method for the chain.
    selections_idx, selections_mask = self._get_indices(chain_id)
    select_unique_idx = [onp.unique(batch_idx, return_inverse=True)
                     for batch_idx in selections_idx]

    # Slice the data and transform into pytree
    selected_observations = []
    for leaf_name in self._reference_data:
      unique_selections = [jnp.array(self._dataset[leaf_name][batch_idx])
                           for batch_idx, select_unique in select_unique_idx]
      selected_observations.append([unique[restore_idx]
                                    for unique, (_, restore_idx)
                                    in zip(unique_selections, select_unique_idx)])

    selected_observations = [jnp.array(leaf) for leaf in selected_observations]
    selected_observations = jax.tree_unflatten(self._pytree_structure,
                                               selected_observations)

    return selected_observations, jnp.array(selections_mask, dtype=jnp.bool_)

  def save_state(self, chain_id: int) -> PyTree:
    raise NotImplementedError("Saving of the DataLoader state is not supported.")

  def load_state(self, chain_id: int, data) -> None:
    raise NotImplementedError("Loading of the DataLoader state is not supported.")

  @property
  def _format(self):
    """Returns shape and dtype of a single observation. """
    return self._sample_format

  @property
  def static_information(self):
    """Returns information about total samples count and batch size. """
    information = {
      "observation_count": self._observation_count
    }
    return information

  def close(self):
    self._dataset.close()



"""Testing utility."""
from functools import partial

import numpy as onp
from numpy import testing

from jax import tree_util

assert_equal = partial(tree_util.tree_map, testing.assert_array_equal)


def assert_close(x, y, **kwargs):
  if "rtol" not in kwargs.keys():
    kwargs["rtol"] = 1e-5
  def assert_fn(xi, yi):
    xi = onp.ravel(xi)
    yi = onp.ravel(yi)
    testing.assert_allclose(xi, yi, **kwargs)
  tree_util.tree_map(assert_fn, x, y)



# Copyright 2021 Multiscale Modeling of Fluid Materials, TU Munich
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Data input in jit-compiled functions

Limited device memory disallows to store all reference data on the device for
big datasets. With the following functions, mini-batches of data can be
requested in ``jit``-compiled functions without loading the entire dataset into
the device's memory.

In the background, a cache of mini-batches is sequentially requested by the
Host Callback Wrappers from the Data Loaders and loaded on the device via the
:mod:`jax_sgmc.util.host_callback` module.

"""

import abc
import threading

import warnings

from functools import partial
import itertools

from typing import Tuple, Any, Callable, Union, NamedTuple, Dict, Protocol, List

import jax
from jax import tree_util, lax
import jax.numpy as jnp
from jax.experimental import host_callback as hcb

import numpy as onp


from jax_sgmc.util import Array, stop_vmap
from jax_sgmc.util.uuid import JaxUUID

class MiniBatchInformation(NamedTuple):
  """Bundles all information about the reference data.

  Args:
    observation_count: Total number of observations
    effective_observation_count: The number of observations without the
      discarded samples remaining after e.g. shuffling.
    mini_batch: List of tuples, tuples consist of ``(observations, parameters)``

  """
  observation_count: Array
  mask: Array
  batch_size: Array


PyTree = Any
MiniBatch = Union[Tuple[PyTree],
                  Tuple[PyTree, MiniBatchInformation],
                  Tuple[PyTree, MiniBatchInformation, Array]]

class DataLoader(metaclass=abc.ABCMeta):
  """Abstract class to define required methods of a DataLoader.

  This class defines common methods of a DataLoader, such as returning an all-
  zero batch with correct shape to initialize the model.
  """

  @property
  @abc.abstractmethod
  def static_information(self) -> Dict:
    """Information about the dataset such as the total observation count. """

  @property
  @abc.abstractmethod
  def _format(self):
    """dtype and shape of a single sample."""

  def initializer_batch(self, mb_size: int = None) -> PyTree:
    """Returns a zero-like mini-batch.

    Args:
      mb_size: Number of observations in a batch. If ``None``, the returned
        pytree has the shape of a single observation.

    """
    obs_format = self._format

    # Append the cache size to the batch_format
    def append_cache_size(leaf):
      if mb_size is None:
        new_shape = tuple(int(s) for s in leaf.shape)
      else:
        new_shape = tuple(int(s) for s in
                          itertools.chain([mb_size], leaf.shape))
      return jnp.zeros(
        dtype=leaf.dtype,
        shape=new_shape
      )

    batch = tree_util.tree_map(append_cache_size, obs_format)
    return batch


class DeviceDataLoader(DataLoader, metaclass=abc.ABCMeta):
  """Abstract class to define required methods of a DeviceDataLoader.

  A class implementing the data loader must have the functionality to return the
  complete dataset as a dictionary of arrays.
  """

  @abc.abstractmethod
  def init_random_data(self, *args, **kwargs) -> PyTree:
    """Initializes the state necessary to randomly draw data. """

  @abc.abstractmethod
  def get_random_data(self,
                      state,
                      batch_size
                      ) ->Tuple[PyTree, Tuple[PyTree, MiniBatchInformation]]:
    """Returns a random batch of the data.

    This function must be jit-able and free of side effects.

    """

  @abc.abstractmethod
  def get_full_data(self) -> Dict:
    """Returns the whole dataset as dictionary of arrays."""

class HostDataLoader(DataLoader, metaclass=abc.ABCMeta):
  """Abstract class to define required methods of a HostDataLoader.

  A class implementing the data loader must have the functionality to load data
  from storage in an ordered and a random fashion.
  """

  def save_state(self, chain_id: int):
    """Returns all necessary information to restore the dataloader state.

    Args:
      chain_id: Each chain can be checkpointed independently.

    Returns:
      Returns necessary information to restore the state of the chain via
      :func:`load_state`.

    """
    raise NotImplementedError("This method must be overwritten to allow "
                              "checkpointing of the data loader.")

  def load_state(self, chain_id: int, data):
    """Restores dataloader state from previously computed checkpoint.

    Args:
      chain_id: The chain to restore the state.
      data: Data from :func:`save_state` to restore state of the chain.

    """
    raise NotImplementedError("This method must be overwritten to allow "
                              "checkpointing of the data loader.")

  @abc.abstractmethod
  def register_random_pipeline(self,
                               cache_size: int = 1,
                               mb_size: int = None,
                               **kwargs
                               ) -> int:
    """Register a new chain which assembles batches randomly.

    Args:
      cache_size: The number of drawn batches.
      mb_size: The number of observations per batch.
      seed: Set the random seed to start the chain at a well-defined state.

    Returns:
      Returns the id of the new chain.

    """

  @abc.abstractmethod
  def register_ordered_pipeline(self,
                                cache_size: int = 1,
                                mb_size: int = None,
                                **kwargs
                                ) -> int:
    """Register a chain which assembles batches in an ordered manner.

    Args:
      cache_size: The number of drawn batches.
      mb_size: The number of observations per batch.

    Returns:
      Returns the id of the new chain.

    """


  @abc.abstractmethod
  def get_batches(self, chain_id: int) -> Tuple[PyTree, Union[Array, None]]:
    """Return batches from an ordered or random chain. """

  def batch_format(self,
                   cache_size: int,
                   mb_size: int,
                   ) -> Tuple[PyTree, MiniBatchInformation]:
    """Returns dtype and shape of cached mini-batches.

    Args:
      cache_size: number of cached mini-batches

    Returns:
      Returns a pytree with the same tree structure as the random data cache but
      with ``jax.ShapedDtypeStruct`` as leaves.
    """
    obs_format = self._format

    # Append the cache size to the batch_format
    def append_cache_size(leaf):
      new_shape = tuple(int(s) for s in
                        itertools.chain([cache_size, mb_size], leaf.shape))
      return jax.ShapeDtypeStruct(
        dtype=leaf.dtype,
        shape=new_shape
      )
    format = tree_util.tree_map(append_cache_size, obs_format)
    mb_info = MiniBatchInformation(
      observation_count=self.static_information["observation_count"],
      batch_size=mb_size,
      mask=onp.ones(mb_size, dtype=onp.bool_))
    return format, mb_info

class CacheState(NamedTuple):
  """Caches several batches of randomly batched reference data.

  Args:
    cached_batches: An array of mini-batches
    cached_batches_count: Number of cached mini-batches. Equals the first
      dimension of the cached batches
    current_line: Marks the next batch to be returned.
    chain_id: Identifier of the chain
    state: Additional information
    valid: Array containing information about the validity of individual samples
  """
  callback_uuid: JaxUUID = None
  cached_batches: PyTree = None
  cached_batches_count: Array = None
  current_line: Array = None
  chain_id: Array = None
  state: PyTree = None
  valid: Array = None
  token: JaxUUID = None

random_data_state = CacheState

Batch = Union[Tuple[CacheState, PyTree],
              Tuple[CacheState, Tuple[PyTree, MiniBatchInformation]]]

class GetBatchFunction(Protocol):
  def __call__(self,
               data_state: CacheState,
               information: bool = False,
               device_count: int = 1) -> Batch:
    """Draws a batch of data.

    Args:
      data_state: State of the chain containing id and cached batches
      information: Include namedtuple containing information about the data
        and batch
      device_count: Number of the devices on which this function is going to be
        called with replicated data states.

    Returns:
      Returns the new state of the random chain and a batch. Optionally a
      namedtuple containing information about the batch and dataset can be
      returned.

    """

class MaskedMappedFunction(Protocol):
  def __call__(self,
               batch: PyTree,
               mask: Array,
               state: PyTree
               ) -> Tuple[PyTree, PyTree]:
    """Function which can be mapped over the whole dataset.

    A function of this form must be passed to the full data map function if it
    is called with ``masking = True``.

    Args:
      batch: Batch of data
      mask: Array marking invalid (double) samples
      state: Variables which results are used in the next computation

    Returns:
      Must return a tuple consisting of the computation results and the state
      which should be used in the computation of the next batch.

    """

class UnmaskedMappedFunction(Protocol):
  def __call__(self,
               batch: PyTree,
               state: PyTree
               ) -> Tuple[PyTree, PyTree]:
    """Function which can be mapped over the whole dataset.

    A function of this form must be passed to the full data map function if it
    is called with ``masking = False``.

    Args:
      batch: Batch of data
      state: Variables which results are used in the next computation

    Returns:
      Must return a tuple consisting of the computation results and the state
      which should be used in the computation of the next batch.

    """

MappedFunction = Union[MaskedMappedFunction, UnmaskedMappedFunction]

class FullDataMapFunction(Protocol):
  def __call__(self,
               fun: MappedFunction,
               data_state: CacheState,
               carry: PyTree,
               masking: bool = False,
               information: bool = False,
               device_count: int = 1
               ) -> Tuple[PyTree, PyTree]:
    """Maps a function over the complete dataset.

    Args:
      fun: Function to be mapped over the dataset
      data_state: Namedtuple containing the id of the chain and cached batches
      carry: Variables which are carried over to the next evaluation of ``fun``
      masking: If true, an array marking invalid samples is passed to the
        function such that a single result for a batch of data can be
        calculated. If false, then a result for each observation must be
        returned and the invalid results are discarded after the computation.
      information: Pass the batch information together with the batch
      device_count: Number of the devices on which this function is going to be
        called with replicated data states.

    Returns:
      Returns the new data state and the results of the computation including
      the carry of the last computation:

      ::

        (data_state, (results, carry)) = full_data_map(...)

    """

class FullDataMapperFunction(Protocol):
  def __call__(self,
               fun: MappedFunction,
               carry: PyTree,
               masking: bool = False,
               information: bool = False,
               batched: bool = True,
               device_count: int = 1
               ) -> PyTree:
    """Maps a function over the complete dataset.

    This function differs to :class:`FullDataMapFunction` that it acquires a
    :class:`CacheState` before each mapping over the full dataset.

    Args:
      fun: Function to be mapped over the dataset
      carry: Variables which are carried over to the next evaluation of ``fun``
      masking: If true, an array marking invalid samples is passed to the
        function such that a single result for a batch of data can be
        calculated. If false, then a result for each observation must be
        returned and the invalid results are discarded automatically after the
        computation.
      information: Pass the batch information together with the batch
      batched: Whether function to be mapped over full dataset is vectorized. If
        false, the function is vmapped such that it can process a batch of
        observations.
      device_count: Number of the devices on which this function is going to be
        called with replicated data states.

    Returns:
      Returns the results of the computation including the carry of the last
      computation:

      ::

        (results, carry) = full_data_mapper(...)

    """

RandomBatch = Tuple[Any, GetBatchFunction, Callable[[], None]]
OrderedBatch = Tuple[Any, FullDataMapFunction, Callable[[], None]]

class _Requests:

  def __init__(self):
    self._cached_requests: Dict[Any, List[int, CacheState, JaxUUID]] = {}
    self._requests: Dict[Any, Dict[int, CacheState, JaxUUID]] = {}
    self._host_data_loaders: Dict[Any, HostDataLoader] = {}

    self._lock = threading.Lock()

  def __call__(self,
               chain_id: int,
               token: JaxUUID,
               device: jax.Device,
               callback_uuid: JaxUUID,
               device_count: int = 1,
               strict: bool = False):
    with self._lock:
      if callback_uuid.as_uuid not in self._requests.keys():
        self._requests[callback_uuid.as_uuid] = {}

      # The token that is expected for a new callback
      current_token = self._requests[callback_uuid.as_uuid].get(int(chain_id))

      # Check if token is from previous request
      old_request = self._cached_requests.get(token.as_uuid)
      if old_request is not None:
        counter, callback_response, new_token = old_request
        if counter == 1:
          # The callback response is no longer necessary
          del self._cached_requests[token.as_uuid]
        else:
          self._cached_requests[token.as_uuid][0] -= 1

      # Check if token is invalid
      elif current_token != token.as_uuid and current_token is not None:
        if strict:
          raise RuntimeError(f"Device {device} made an invalid request for "
                             f"chain {chain_id}. This might be due to using a "
                             f"pmap in a jitted function. See "
                             f"usage/data.html#combining-pmap-and-jit in the "
                             f"docs.")
        warnings.warn(f"Device {device} made an invalid request for chain "
                      f"{chain_id}. This might be due to using a pmap in a "
                      f"jitted function. If the preservation ")

      # The request is valid and the first request, so the results have to be
      # stored in the cache
      else:
        # Issue a new token and request data
        new_token = JaxUUID()
        callback_response = self._host_data_loaders[callback_uuid.as_uuid].get_batches(chain_id)

        self._requests[callback_uuid.as_uuid][int(chain_id)] = new_token.as_uuid
        # Store data if other devices are going to request it
        if device_count != 1:
          self._cached_requests[token.as_uuid] = [
            device_count - 1, callback_response, new_token]

    return callback_response, new_token

  def __getstate__(self):
    state = self.__dict__.copy()
    # The lock should always be reinitialized
    del state['_lock']
    return state

  def __setstate__(self, state):
    self.__dict__.update(state)
    self._lock = threading.Lock()

_data_requests = _Requests()

def random_reference_data(data_loader: DataLoader,
                          cached_batches_count: int,
                          mb_size: int,
                          verify_calls: bool = False
                          ) -> RandomBatch:
  """Initializes reference data access in jit-compiled functions.

  Randomly draw batches from a given dataset on the host or the device.

  Args:
    data_loader: Reads data from storage.
    cached_batches_count: Number of batches in the cache. A larger number is
      faster, but requires more memory.
    mb_size: Size of the data batch.
    verify_calls: Verify calls to the host when using pmap. This might not be
      necessary if batches are assembles randomly without extra conditions.

  Returns:
    Returns a tuple of functions to initialize a new reference data state, get
    a minibatch from the reference data state and release the data loader after
    the last computation.

  """
  # Check batch size is not bigger than total observation count
  observation_count = data_loader.static_information["observation_count"]
  if observation_count < mb_size:
    raise ValueError(f"Batch size cannot be bigger than the number of total "
                     f"observations. Got {observation_count} and {mb_size}.")
  if cached_batches_count <= 0 or mb_size <= 0:
    raise ValueError(f"Cache size and batch size must be positive, got"
                     f"{cached_batches_count} and {mb_size}.")

  if isinstance(data_loader, HostDataLoader):
    return _random_reference_data_host(
      data_loader, cached_batches_count, mb_size,
      verify_calls=verify_calls)
  elif isinstance(data_loader, DeviceDataLoader):
    if not cached_batches_count == 1:
      raise ValueError("No caching on device.")
    return _random_reference_data_device(
      data_loader, mb_size)
  else:
    raise TypeError("The DataLoader must inherit from HostDataLoader or "
                    "DeviceDataLoader")

def full_reference_data(data_loader: DataLoader,
                        cached_batches_count: int = 100,
                        mb_size: int = None
                        ) -> OrderedBatch:
  """Initializes reference data access in jit-compiled functions.

  Map a function batch-wise over a dataset on the host or the device.

  Args:
    data_loader: Reads data from storage.
    cached_batches_count: Number of batches in the cache. A larger number is
      faster, but requires more memory.
    mb_size: Size of the data batch.

  Returns:
    Returns a tuple of functions to initialize a new reference data state, map a
    function over the complete dataset and release the data loader after the
    last computation.

  """
  # Check batch size is not bigger than total observation count
  observation_count = data_loader.static_information["observation_count"]
  if observation_count < mb_size:
    raise ValueError(f"Batch size cannot be bigger than the number of total "
                     f"observations. Got {observation_count} and {mb_size}.")

  if isinstance(data_loader, HostDataLoader):
    init_fn, (_batch_fn, mb_information), cleanup = _full_reference_data_host(
      data_loader, cached_batches_count, mb_size)
  elif isinstance(data_loader, DeviceDataLoader):
    init_fn, (_batch_fn, mb_information), cleanup = _full_reference_data_device(
      data_loader, cached_batches_count)
  else:
    raise TypeError("The DataLoader must inherit from HostDataLoader or "
                    "DeviceDataLoader")

  num_iterations = int(onp.ceil(
    mb_information.observation_count / mb_information.batch_size))

  batch_size = mb_information.batch_size

  def _uninitialized_body_fn(fun,
                             state,
                             iteration,
                             information=False,
                             masking=False,
                             device_count=1):
    # The mask has is 1 if the observation is valid and 0 otherwise. This is
    # necessary to ensure, that fun is always called with the same tree shape.
    observations = iteration * batch_size + jnp.arange(batch_size)
    mask = observations < mb_information.observation_count

    data_state, fun_state = state
    data_state, batch = _batch_fn(
      data_state,
      information=information, device_count=device_count)

    if masking:
      result, fun_state = fun(batch, mask, fun_state)
    else:
      result, fun_state = fun(batch, fun_state)

    return (data_state, fun_state), result

  def batch_scan(fun: MappedFunction,
                 data_state: CacheState,
                 carry: PyTree,
                 masking: bool = False,
                 information: bool = False,
                 device_count: int = 1
                 ) -> Tuple[PyTree, PyTree]:
    """Maps the function over all data.

    Args:
      fun: Function accepting a batch of data, a mask and a state.
      data_state: Reference data state.
      carry: A argument that is carried over between iterations
      masking: If set to true, the mapped function is called with a positional
        argument mask and expected to return the results with a reduced dimension.
        Setting to true changes the signature from `fun(data, carry)` to
        `fun(data, mask, carry)`.
      information: Provide the minibatch information in addition to the data
        batch.

    """
    _body_fn = partial(_uninitialized_body_fn,
                       fun,
                       information=information,
                       masking=masking,
                       device_count=device_count)
    (data_state, carry), results = lax.scan(
      _body_fn, (data_state, carry), onp.arange(num_iterations))

    if masking:
      true_results = results
    else:
      # The results must be concatenated
      concat_results = tree_util.tree_map(
        partial(jnp.concatenate, axis=0),
        results)
      # Invalid results (fillers) must be thrown away
      true_results = tree_util.tree_map(
        lambda leaf: leaf[0:mb_information.observation_count],
        concat_results)

    return data_state, (true_results, carry)

  return init_fn, batch_scan, cleanup


def tree_dtype_struct(pytree: PyTree):
  """Returns a tree with leaves only representing shape and type."""
  @partial(partial, tree_util.tree_map)
  def concrete_to_shape_struct(leaf):
    shape_struct = jax.ShapeDtypeStruct(
      dtype=leaf.dtype,
      shape=leaf.shape)
    return shape_struct
  return concrete_to_shape_struct(pytree)

def tree_index(pytree: PyTree, index):
  """Indexes the leaves of the tree in the first dimension.

  Args:
    pytree: Tree to index with array-like leaves
    index: Selects which slice to return

  Returns:
    Returns a tree with the same structure as pytree, but the leaves have a
    dimension reduced by 1.

  """
  @partial(partial, tree_util.tree_map)
  def split_tree_imp(leaf):
    if leaf.ndim == 1:
      return leaf[index]
    else:
      return leaf[index, ::]
  return split_tree_imp(pytree)


# Callback is independent of assembling of the batches
def _hcb_wrapper(data_loader: HostDataLoader,
                 cached_batches_count: int,
                 mb_size: int,
                 verify_calls: bool = False
                 ) -> Tuple[GetBatchFunction, MiniBatchInformation]:
  # These are helper function which keep a reference to the stateful data object
  # and can be called via the host_callback.call function
  # The format of the mini batch is static.

  hcb_format, mb_information = data_loader.batch_format(
    cached_batches_count, mb_size=mb_size)
  mask_shape = (cached_batches_count, mb_size)

  # The definition requires passing an argument to the host function. The shape
  # of the returned data must be known before the first call. The chain id
  # determines whether the data is collected randomly or sequentially.

  def get_data(req, device):
    chain_id, callback_uuid, token, device_count = req
    # The data request class takes care of assigning the request to the right
    # data loader and verifies it.
    (new_data, mask), new_token = _data_requests(
      chain_id, token, device, callback_uuid, device_count,
      strict=verify_calls)
    if mask is None:
      # Assume all samples to be valid. It is important to perform the creation
      # of the array on the host, as otherwise a deadlock will occur.
      mask = onp.ones(mask_shape, dtype=jnp.bool_)
    return new_data, mask, new_token

  def _new_cache_fn(req: Tuple[CacheState, int]) -> CacheState:
    """This function is called if the cache must be refreshed."""
    state, device_count = req
    new_data, masks, token = hcb.call(
      get_data,
      (state.chain_id, state.callback_uuid, state.token, device_count),
      result_shape=(
        hcb_format,
        jax.ShapeDtypeStruct(shape=mask_shape, dtype=jnp.bool_),
        JaxUUID()),
      call_with_device=True)
    new_state = CacheState(
      cached_batches_count=state.cached_batches_count,
      cached_batches=new_data,
      current_line=jnp.array(0),
      chain_id=state.chain_id,
      valid=masks,
      callback_uuid=state.callback_uuid,
      token=token)
    return new_state

  def _old_cache_fn(req: Tuple[CacheState, int]) -> CacheState:
    """This function is called if the cache must not be refreshed."""
    return req[0]

  # Necessary, because cond is replaced by select under vmap, but the cond
  # branches have side effects.
  @stop_vmap.stop_vmap
  def _data_state_helper(data_state, device_count):
    return lax.cond(data_state.current_line == data_state.cached_batches_count,
                    _new_cache_fn,
                    _old_cache_fn,
                    (data_state, device_count))

  def batch_fn(data_state: CacheState,
               information: bool = False,
               device_count: int = 1
               ) -> Batch:
    """Draws a new random batch (hides data transfer between devices).

    Args:
      data_state: State with cached samples
      information: Whether to return batch information
      device_count: Number of parallel programs calling the batch function

    Returns:
      Returns the new data state and the next batch. Optionally an additional
      struct containing information about the batch can be returned.

    """
    if device_count > jax.device_count():
      raise ValueError(f"The value of device_count cannot exceed the true "
                       f"device count. Expecting device_count (given: "
                       f"{device_count}) <= {jax.device_count()}.")
    if device_count != 1:
      warnings.warn("Changing the device count can cause memory accumulation. "
                    "Continue only if you know what you are doing.")

    # Refresh the cache if necessary, after all cached batches have been used.
    data_state = _data_state_helper(data_state, device_count)
    current_line = jnp.mod(data_state.current_line,
                           data_state.cached_batches_count)

    # Read the current line from the cache and add the mask containing
    # information about the validity of the individual samples
    mini_batch = tree_index(data_state.cached_batches, current_line)
    mask = data_state.valid[current_line, :]

    current_line = current_line + 1

    new_state = CacheState(
      cached_batches=data_state.cached_batches,
      cached_batches_count=data_state.cached_batches_count,
      current_line=current_line,
      chain_id=data_state.chain_id,
      valid=data_state.valid,
      callback_uuid=data_state.callback_uuid,
      token=data_state.token)

    info = MiniBatchInformation(
      observation_count = mb_information.observation_count,
      batch_size = mb_information.batch_size,
      mask = mask)

    if information:
      return new_state, (mini_batch, info)
    else:
      return new_state, mini_batch

  return batch_fn, mb_information


def _random_reference_data_host(data_loader: HostDataLoader,
                                cached_batches_count: int = 100,
                                mb_size: int = 1,
                                verify_calls: bool = False) -> RandomBatch:
  """Random reference data access via host-callback. """
  # Warn if cached_batches are bigger than total dataset
  observation_count = data_loader.static_information["observation_count"]
  if observation_count < cached_batches_count * mb_size:
    warnings.warn("Cached batches are bigger than the total dataset. Consider "
                  "using a DeviceDataLoader.")

  batch_fn, _ = _hcb_wrapper(
    data_loader, cached_batches_count, mb_size,
    verify_calls=verify_calls)

  callback_uuid = JaxUUID()
  _data_requests._host_data_loaders[callback_uuid.as_uuid] = data_loader

  def init_fn(**kwargs) -> CacheState:
    # Pass the data loader the information about the number of cached
    # mini-batches. The data loader returns an unique id for reproducibility
    chain_id = data_loader.register_random_pipeline(
      cached_batches_count,
      mb_size=mb_size,
      **kwargs)
    initial_state, initial_mask = _data_requests._host_data_loaders[callback_uuid.as_uuid].get_batches(chain_id)
    if initial_mask is None:
      initial_mask = jnp.ones((cached_batches_count, mb_size), dtype=jnp.bool_)
    inital_cache_state = CacheState(
      cached_batches=initial_state,
      cached_batches_count=jnp.array(cached_batches_count),
      current_line=jnp.array(0),
      chain_id=jnp.array(chain_id),
      valid=initial_mask,
      callback_uuid=callback_uuid,
      token=JaxUUID())
    return inital_cache_state

  def release():
    del _data_requests._host_data_loaders[callback_uuid.as_uuid]

  return init_fn, batch_fn, release


def _random_reference_data_device(data_loader: DeviceDataLoader,
                                  mb_size: int
                                  ) -> RandomBatch:
  """Random reference data on device. """

  def init_fn(**kwargs) -> CacheState:
    state = data_loader.init_random_data(**kwargs)
    return CacheState(state=state)

  def batch_fn(state: CacheState,
               information: bool = False
               ) -> Batch:

    state, (batch, mb_info) = data_loader.get_random_data(
      state.state, batch_size=mb_size)
    new_state = CacheState(state=state)

    if information:
      return new_state, (batch, mb_info)
    else:
      return new_state, batch

  def release():
    pass

  return init_fn, batch_fn, release

def _full_reference_data_host(data_loader: HostDataLoader,
                              cached_batches_count: int = 100,
                              mb_size: int = None
                              ) -> Tuple[Callable, Tuple[Callable, MiniBatchInformation], Callable]:
  """Sequentially load batches of reference data via host-callback. """
  # Warn if cached_batches are bigger than total dataset
  observation_count = data_loader.static_information["observation_count"]
  if observation_count < cached_batches_count * mb_size:
    warnings.warn("Cached batches are bigger than the total dataset. Consider "
                  "using a DeviceDataLoader.")

  batch_fn = _hcb_wrapper(
    data_loader,
    cached_batches_count,
    mb_size,
    verify_calls=True)

  # Register the data loader
  callback_uuid = JaxUUID()
  _data_requests._host_data_loaders[callback_uuid.as_uuid] = data_loader

  def init_fn(**kwargs) -> CacheState:
    # Pass the data loader the information about the number of cached
    # mini-batches. The data loader returns an unique id for reproducibility
    chain_id = _data_requests._host_data_loaders[callback_uuid.as_uuid].register_ordered_pipeline(
      cached_batches_count,
      mb_size=mb_size,
      **kwargs)
    initial_state, initial_mask = _data_requests._host_data_loaders[callback_uuid.as_uuid].get_batches(chain_id)
    if initial_mask is None:
      initial_mask = jnp.ones((cached_batches_count, mb_size), dtype=jnp.bool_)
    inital_cache_state=CacheState(
      cached_batches=initial_state,
      cached_batches_count=jnp.array(cached_batches_count),
      current_line=jnp.array(0),
      chain_id=jnp.array(chain_id),
      valid=initial_mask,
      callback_uuid=callback_uuid,
      token=JaxUUID())
    return inital_cache_state

  def release():
    del _data_requests._host_data_loaders[callback_uuid.as_uuid]

  return init_fn, batch_fn, release


def _full_reference_data_device(data_loader: DeviceDataLoader,
                                mb_size: int = None
                                ) -> Tuple[Callable,
                                           Tuple[Callable,
                                                 MiniBatchInformation],
                                           Callable]:
  """Batches the dataset on the device. """

  reference_data = data_loader.get_full_data()
  total_observations = data_loader.static_information["observation_count"]


  # The information about the batches need to be static.
  mb_info = MiniBatchInformation(
    observation_count=total_observations,
    batch_size=mb_size)

  def init_fn(offset: jnp.ndarray = 0):
    if offset >= total_observations:
      raise ValueError(f"The offset cannot be greater than the total "
                       f"observation count. Given {offset} and "
                       f"{total_observations}.")

    init_state = CacheState(current_line=offset)
    return init_state

  def batch_fn(data_state: CacheState,
               information: bool = False):
    indices = jnp.mod(jnp.arange(mb_size) + data_state.current_line,
                      total_observations)

    # Update the offset, where to start slicing in the next iteration.
    new_state = CacheState(
      current_line=jnp.mod(indices[-1] + 1, total_observations))

    selected_data = tree_index(reference_data, indices)

    if information:
      return new_state, (selected_data, mb_info)
    else:
      return new_state, selected_data

  def release():
    pass

  return init_fn, (batch_fn, mb_info), release


class _FullDataHelper:
  """Class to keep track of unused CacheStates. """

  def __init__(self, data_loader, cache_size, batch_size):
    self._init_fn, self._map_fn, self._cleanup_fn = full_reference_data(
      data_loader, cache_size, batch_size)

    # Initialize the first cache state to compute the shape of the CacheState
    self._unused_states = [self._new_cache_state()]
    self._cache_state_format = tree_util.tree_map(
      lambda leaf: jax.ShapeDtypeStruct(dtype=leaf.dtype, shape=leaf.shape),
      self._unused_states[0])

    self._lock = threading.Lock()

  @property
  def get_map_fn(self) -> FullDataMapFunction:
    return self._map_fn

  @property
  def get_cache_state_format(self) -> PyTree:
    return self._cache_state_format

  def get_cache_state(self) -> CacheState:
    """Return unused cache state or create new one if none available."""
    with self._lock:
      if len(self._unused_states) == 0:
        state = self._new_cache_state()
      else:
        state = self._unused_states.pop()
    return state

  def free_cache_state(self, cache_state: CacheState) -> Array:
    """Adds cache state back to the unused cache states."""
    with self._lock:
      self._unused_states.append(cache_state)
    return jnp.array(1.0)

  def _new_cache_state(self) -> CacheState:
    """Creates a new cache state. """
    new_cache_state = self._init_fn()
    return new_cache_state

  def cleanup(self):
    self._cleanup_fn()
    self._unused_states = None

  def __getstate__(self):
    state = self.__dict__.copy()
    del state["_lock"]
    return state

  def __setstate__(self, state):
    self.__dict__.update(state)
    self._lock = threading.Lock()


def full_data_mapper(data_loader: DataLoader = None,
                     cached_batches_count: int = 1,
                     mb_size: int = 1
                     ) -> Tuple[FullDataMapperFunction, Callable]:
  """Initializes a functional to map a function over a complete dataset.

  This function extends the functionality of
  :func:`full_reference_data` by loading the data states
  form the host before each mapping.

  Args:
    data_loader: Reads data from storage
    cached_batches_count: Number of batches in the cache. A larger number is
      faster, but requires more memory
    mb_size: Size of the data batch

  Returns:
    Returns a tuple of functions to map another function over a complete dataset
    of an appropriate :class:`DataLoader` and another function to release
    the data loader after the last computation.

  """

  _helper = _FullDataHelper(data_loader, cached_batches_count, mb_size)

  # Helper functions to load/save the CacheStates on the host via host_callback
  def _get_state() -> CacheState:
    cache_state = hcb.call(
      lambda _: _helper.get_cache_state(),
      jnp.array(1.0),
      result_shape=_helper.get_cache_state_format)
    return cache_state

  def _free_cache_state(cache_state: CacheState, results: PyTree) -> Array:
    # Loop-through the results to hinder XLA to remove the tap call
    results = hcb.id_tap(
      lambda cs, _: _helper.free_cache_state(cs),
      cache_state,
      result=results)
    return results

  def mapper_fn(fun: Union[MaskedMappedFunction, UnmaskedMappedFunction],
                carry: PyTree,
                masking: bool = False,
                information: bool = False,
                batched: bool = True,
                device_count: int = 1) -> PyTree:
    data_state = _get_state()

    # Batch the function if it is not batched.
    if batched:
      batched_fun = fun
    else:
      if masking:
        raise ValueError("The function must be vectorized manually to allow "
                         "masking.")

      if mb_size == 1:
        # No vmapping required but first axis of all observations must be
        # removed.
        def batched_fun(batch, state):
          squeezed_batch = tree_util.tree_map(
            partial(jnp.squeeze, axis=0),
            batch)
          result, state = fun(squeezed_batch, state)
          expanded_result = tree_util.tree_map(
            partial(jnp.expand_dims, axis=0),
                    result)
          return expanded_result, state
      else:
        # Only the first resulting state is passed to the next iteration
        vmapped_fun = jax.vmap(fun, in_axes=(0, None))
        def batched_fun(batch, state):
          results, states = vmapped_fun(batch, state)
          if states is None:
            state = None
          else:
            state = tree_util.tree_map(
              partial(jnp.take_along_axis, indices=0, axis=0),
              states)
          return results, state

    (new_data_state, results) = _helper.get_map_fn(
      batched_fun, data_state, carry,
      masking=masking, information=information, device_count=device_count)
    results = _free_cache_state(new_data_state, results)
    return results

  def release():
    _helper.cleanup()

  return mapper_fn, release



# Copyright 2021 Multiscale Modeling of Fluid Materials, TU Munich
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Save and checkpoint chains.

  **JaxSGMC** supports saving and checkpointing inside jit-compiled
  functions. Saving works by combining a Data Collector with the host callback
  wrappers.

"""

# Todo: If dataloader supports checkpointing, it must be able to checkpoint
#       itself.

import abc
import itertools

from pathlib import Path
from functools import partial
import threading

from typing import Any, Union, Tuple, Callable, Type, Dict, NoReturn

from collections import namedtuple

import numpy as onp

import jax.numpy as jnp
from jax import tree_util, lax
from jax.experimental import host_callback

try:
  import ujson
except ModuleNotFoundError:
  ujson = None

try:
  import h5py
  HDF5File = h5py.File
except ModuleNotFoundError:
  HDF5File = None
  h5py = None

# Import haiku to register a rule for transforming FlatMapping into a dict
try:
  import haiku._src.data_structures as haiku_ds
except ModuleNotFoundError:
  haiku_ds = None

from jax_sgmc import data
from jax_sgmc import scheduler
from jax_sgmc.util import stop_vmap

PyTree = Any

# Global rules for translating a tree-node into a dict
_dictionize_rules: Dict[Type, Callable] = {}

def register_dictionize_rule(type: Type) -> Callable[[Callable], None]:
  """Decorator to define new rules transforming a pytree node to a dict.

  By default, transformations are defined for some default types:

    - list
    - dict
    - (named)tuple

  Additionally, transformation for the following optional libraries are
  implemented:

  - haiku._src.data_structures.FlatMapping

  Args:
    type: Type (or class) of the currently undefined node

  Returns:
    The decorated function is not intended to be used directly.

  """
  def register_function(rule: Callable):
    global _dictionize_rules
    assert type not in _dictionize_rules.keys(), f"Rule for {type.__name__} is " \
                                                 f"already defined."
    _dictionize_rules[type] = rule
  return register_function

def _default_dictionize(node):
  leaves = tree_util.tree_leaves(node)
  leaf_names = [f"unknown~pytree~leaf~{idx}" for idx in range(len(leaves))]
  return zip(leaf_names, leaves)

def _dictionize(node):
  """Apply the registered rules to a pytree node."""
  global _dictionize_rules
  # If there are no leaves, then a dictionize rule is not required
  if len(tree_util.tree_leaves(node)) == 0:
    return []
  for node_type, node_rule in _dictionize_rules.items():
    if isinstance(node, node_type):
      return node_rule(node)

  return _default_dictionize(node)

def pytree_to_dict(tree: PyTree):
  """Constructs a dictionary from a pytree.

  Transforms each node of a pytree to a dict by appling defined dictionize. New
  rules can be specified by the :func:`jax_sgmc.io.register_dictionize_rule`
  -decorator.

  .. doctest::

    >>> from jax_sgmc import io
    >>> some_tree = {'a': 0.0, 'b': {'b1': [0.0, 0.1], 'b2': 0.0}}
    >>> as_dict = io.pytree_to_dict(some_tree)
    >>> print(as_dict)
    {'a': 0.0, 'b': {'b1': {'list_element_0': 0.0, 'list_element_1': 0.1}, 'b2': 0.0}}

  Args:
    tree: All nodes of the tree must either have no children or a registered
      transformation to dict.

  Returns:
    Returns the tree as a dict with similar structure.

  """
  global _dictionize_rules
  pytree_def = tree_util.tree_structure(tree)
  if tree_util.treedef_is_leaf(pytree_def):
    return tree
  else:
    return {key: pytree_to_dict(childs)
            for key, childs
            in _dictionize(tree)}

def pytree_dict_keys(tree: PyTree):
  """Returns a list of keys to acces the leaves of the tree.

  Args:
    tree: Pytree as a dict

  Returns:
    Returns a list of tuples, where each tuple contains the keys to access the
    leaf of the flattened pytree in the unflattened dict.

    For example:

    .. doctest::

      >>> from jax import tree_leaves
      >>> from jax_sgmc import io
      >>> pytree = {"a": [0.0, 1.0], "b": 2.0}
      >>> pytree_as_dict = io.pytree_to_dict(pytree)
      >>> pytree_leaves = tree_leaves(pytree_as_dict)
      >>> pytree_keys = io.pytree_dict_keys(pytree_as_dict)
      >>> print(pytree_leaves)
      [0.0, 1.0, 2.0]
      >>> print(pytree_keys)
      [('a', 'list_element_0'), ('a', 'list_element_1'), ('b',)]

  """
  node = pytree_to_dict(tree)
  leaves, treedef = tree_util.tree_flatten(node)
  idx_tree = tree_util.tree_unflatten(treedef, list(range(len(leaves))))
  key_list = [None] * len(leaves)
  def _recurse(node, path):
    if node is None:
      return
    elif isinstance(node, int):
      key_list[node] = path
    else:
      for key, value in node.items():
        _recurse(value, path + [key])
  _recurse(idx_tree, [])
  return [tuple(key) for key in key_list]

def dict_to_pytree(pytree_as_dict: dict, target: PyTree):
  """Restores the original tree structure given by the target from a dict.

  Restores the pytree as a dict to its original tree structure. This function
  can also operate on subtrees, as long as the subtree (a dict) of the pytree
  as dict matches the subtree of the target dict.

  .. doctest::

    >>> from jax_sgmc import io
    >>> some_tree = {'a': 0.0, 'b': {'b1': [0.0, 0.1], 'b2': 0.0}}
    >>> as_dict = io.pytree_to_dict(some_tree)
    >>> sub_pytree = io.dict_to_pytree(as_dict['b'], some_tree['b'])
    >>> print(sub_pytree)
    {'b1': [0.0, 0.1], 'b2': 0.0}

  Args:
    pytree_as_dict: A pytree which has been transformed to a dict of dicts.
    target: A pytree defining the original tree structure.

  """
  target_dict_keys = pytree_dict_keys(target)
  tree_structure = tree_util.tree_structure(target)
  def _recurse_get(key_list):
    key_list = list(key_list)
    element = pytree_as_dict
    while len(key_list) > 0:
      element = element.get(key_list.pop(0))
    return element
  new_leaves = map(_recurse_get, target_dict_keys)
  return tree_util.tree_unflatten(tree_structure, new_leaves)

@register_dictionize_rule(dict)
def _dict_to_dict(some_dict: dict):
  return some_dict.items()

@register_dictionize_rule(list)
def _list_to_dict(some_list: list):
  return ((f"list_element_{idx}", value) for idx, value in enumerate(some_list))

@register_dictionize_rule(tuple)
def _namedtuple_to_dict(some_tuple: Union[tuple, namedtuple]):
  if hasattr(some_tuple, '_fields'):
    return some_tuple._asdict().items()
  else:
    return ((f"list_element_{idx}", value) for idx, value in enumerate(some_tuple))

if haiku_ds is not None:
  @register_dictionize_rule(haiku_ds.FlatMapping)
  def _flat_mapping_to_dict(flat_mapping: haiku_ds.FlatMapping):
    return haiku_ds.to_immutable_dict(flat_mapping).items()

saving_state = namedtuple("saving_state",
                          ["chain_id",
                           "saved_samples",
                           "data"])

Saving = Tuple[Callable[[PyTree, PyTree, PyTree], saving_state],
               Callable[[saving_state,
                         jnp.bool_,
                         PyTree,
                         PyTree,
                         PyTree],
                        Union[Any, NoReturn]],
               Callable[[Any], Union[Any, NoReturn]]]

class DataCollector(metaclass=abc.ABCMeta):
  """Collects sampled data and data loader states. """

  @abc.abstractmethod
  def register_data_loader(self, data_loader: data.DataLoader):
    """Register data loader to save the state. """

  @abc.abstractmethod
  def register_chain(self,
                     init_sample: PyTree = None,
                     init_checkpoint: PyTree = None,
                     static_information: PyTree = None
                     ) -> int:
    """Registers a chain to save samples from.

    Args:
      init_sample: Determining shape, dtype and tree structure of sample.
      init_checkpoint: Determining shape, dtype and tree structure of solver
        state to enable checkpointing.
      static_information: Information about the total number of collected
        samples

    Returns:
      Returns id of the new chain.
    """

  @abc.abstractmethod
  def save(self, chain_id: int, values):
    """Called with collected samples. """

  @abc.abstractmethod
  def finalize(self, chain_id: int):
    """Called after solver finished. """

  @abc.abstractmethod
  def finished(self, chain_id: int):
    """Called in main thread after jax threads have been released."""

  @abc.abstractmethod
  def checkpoint(self, chain_id: int, state):
    """Called every nth step. """

  @abc.abstractmethod
  def resume(self):
    """Called to restore data loader state and return sample states. """

# Todo: Maybe remove? Not usable for big data -> better hdf5
class JSONCollector(DataCollector):
  """Saves samples in json format.

  Args:
    dir: Directory to save the collected samples.
    write_frequency: Number of samples to be collected before a file is written

  """

  def __init__(self, dir: str, write_frequency=0):
    assert ujson is not None, "ujson is required to save samples to json files"
    self._dir = Path(dir)
    self._dir.mkdir(exist_ok=True)
    self._collected_samples = []
    self._sample_count = []
    self._write_frequency = write_frequency

  def register_data_loader(self, data_loader: data.DataLoader):
    """Not supported for JSON collector. """
    raise NotImplementedError("Checkpointing is not supported by JSON loader.")

  def register_chain(self,
                     init_sample: PyTree = None,
                     init_checkpoint: PyTree = None,
                     static_information: PyTree = None
                     ) -> int:
    """Register a chain to save samples from. """
    new_chain = len(self._collected_samples)
    self._collected_samples.append([])
    self._sample_count.append(0)
    return new_chain

  def save(self, chain_id: int, values):
    """Called with collected samples. """
    # Store new values
    self._collected_samples[chain_id].append(
      tree_util.tree_map(onp.array, values)
    )
    self._sample_count[chain_id] += 1
    # Write to file but keep collected samples in memory
    if self._write_frequency > 0:
      if (self._sample_count[chain_id] % self._write_frequency) == 0:
        self._write_file(chain_id, self._sample_count[chain_id])

  def finalize(self, chain_id: int):
    """Called after solver finished. """
    if self._sample_count[chain_id] > 0:
      self._write_file(chain_id, self._sample_count[chain_id])

  def checkpoint(self, chain_id: int, state):
    """Called every nth step. """
    raise NotImplementedError("Checkpointing is not supported by JSON loader.")

  def resume(self):
    """Called to restore data loader state and return sample states. """
    raise NotImplementedError("Checkpointing is not supported by JSON loader.")

  def _write_file(self, chain_id, iteration):
    filename = self._dir / f"chain_{chain_id}_iteration_{iteration}.json"
    filename.touch()
    # Transform collected samples to list after chaining together
    stacked_samples = tree_util.tree_map(
      lambda *samples: onp.stack(samples, axis=0),
      *self._collected_samples[chain_id]
    )
    samples_as_list = tree_util.tree_map(
      lambda leaf: leaf.tolist(),
      stacked_samples
    )
    with open(filename, "w") as file:
      ujson.dump(samples_as_list, file) # pylint: disable=c-extension-no-member

  def finished(self, chain_id: int):
    """Simply return, nothing to wait for. """
    return None


class HDF5Collector(DataCollector):
  """Save to hdf5 format.

  This data collector supports serializing collected samples and checkpoints
  into the hdf5 file format. The samples are saved in a structure similar to the
  original pytree and can thus be viewed easily via the hdf5-viewer.

  Note:
    This class requires that ``h5py`` is installed. Additional information
    can be found in the :ref:`installation instructions<additional_requirements>`.

  .. doctest::

    >>> import tempfile
    >>> tf = tempfile.TemporaryFile()
    >>>
    >>> import h5py
    >>> from jax_sgmc import io
    >>>
    >>> file = h5py.File(tf, "w") # Use a real file for real saving
    >>> data_collector = io.HDF5Collector(file)
    >>> saving = io.save(data_collector)
    >>>
    >>> # ... use the solver ...
    >>>
    >>> # Close the file
    >>> file.close()

  Args:
    file: hdf5 file object

  """

  def __init__(self, file: HDF5File):
    assert h5py is not None, "h5py must be installed to use this DataCollector."
    # Barrier to wait until all data has been processed
    self._finished = []
    self._sample_count = []
    self._leaf_names = []
    self._file = file

  def register_data_loader(self, data_loader: data.DataLoader):
    """Registers data loader to save the state. """

  def register_chain(self,
                     init_sample: PyTree = None,
                     init_checkpoint: PyTree = None,
                     static_information: PyTree = None
                     ) -> int:
    """Registers a chain to save samples from.

    Args:
      init_sample: Pytree determining tree structure and leafs of samples to be
        saved.
      init_checkpoint: Pytree determining tree structure and leafs of the states
        to be checkpointed
      static_information: Information about the total numbers of samples
        collected.
    """
    assert init_sample is not None, "Need a sample-pytree to allocate memory."
    assert init_checkpoint is not None, "Need a checkpoint-pytree to allocate " \
                                        "memory."

    chain_id = len(self._finished)
    # Save the pytree as it would be transformed to a dict by pytree_to_dict
    leaf_names = ["/".join(itertools.chain([f"/chain~{chain_id}"], key_tuple))
                  for key_tuple in pytree_dict_keys(init_sample)]
    leaves = tree_util.tree_leaves(init_sample)
    # Allocate memory upfront, this is more effective than the chunk-wise
    # allocation
    for leaf_name, leaf in zip(leaf_names, leaves):
      new_shape = tuple(int(s) for s in
                        itertools.chain([static_information.samples_collected],
                                        leaf.shape))
      self._file.create_dataset(
        leaf_name,
        shape=new_shape,
        dtype=leaf.dtype)
    self._sample_count.append(0)
    self._leaf_names.append(leaf_names)
    self._finished.append(threading.Barrier(2))
    return chain_id

  def save(self, chain_id: int, values):
    """Saves new leaves to dataset.

    Args:
      chain_id: ID from register_chain
      values: Tree leaves of sample

    """
    for leaf_name, value in zip(self._leaf_names[chain_id], values):
      self._file[leaf_name][self._sample_count[chain_id]] = value
    self._sample_count[chain_id] += 1

  def finalize(self, chain_id: int):
    """Waits for all writing to be finished (scheduled via host_callback).

    Args:
      chain_id: Id of the chain which is finished

    """
    # Is called after all host callback calls have been processed
    # self._finished[chain_id].wait()

  def checkpoint(self, chain_id: int, state):
    """Called every nth step. """

  def resume(self):
    """Called to restore data loader state and return sample states. """

  def finished(self, chain_id: int):
    """Returns after everything has been written to the file.

    Finalize is scheduled via host_callback and finished is called in the normal
    python flow. Via a barrier it is possible to pause the program flow until
    all asynchronously saved data has been processed.

    Args:
      chain_id: ID of chain requesting continuation of normal program flow.

    """
    # self._finished[chain_id].wait()

class MemoryCollector(DataCollector):
  """Stores samples entirely in RAM (numpy arrays).

  The RAM is usually larger than the device array and thus allows to store a
  greater number of samples.

  Args:
    save_dir: Directory to output results as numpy-npz with one file per chain.
      If none, the results will only be returned.

  """

  def __init__(self, save_dir=None):
    self._finished = []
    self._samples = []
    self._samples_count = []
    self._treedefs = []
    self._leafnames = []
    self._dir = save_dir

  def register_data_loader(self, data_loader: data.DataLoader):
    """Registers data loader to save the state. """

  def register_chain(self,
                     init_sample: PyTree = None,
                     static_information = None,
                     **unused_kwargs) -> int:
    """Registers a chain to save samples from.

    Args:
      init_sample: Pytree determining tree structure and leafs of samples to be
        saved.
      static_information: Information about the total numbers of samples
        collected.
    """
    assert init_sample is not None, "Need a sample-pytree to allocate memory."

    chain_id = len(self._finished)
    leaves, treedef = tree_util.tree_flatten(init_sample)
    def leaf_shape(leaf):
      new_shape = onp.append(
        static_information.samples_collected,
        leaf.shape)
      new_shape = tuple(int(s) for s in new_shape)
      return new_shape
    sample_cache = [onp.zeros(leaf_shape(leaf), dtype=leaf.dtype) for leaf in leaves]

    # Only generate the keys for each leaf if necessary
    if self._dir:
      pytree_keys = pytree_dict_keys(init_sample)
    else:
      pytree_keys = [f"leaf~{idx}" for idx in range(len(leaves))]

    self._finished.append(threading.Lock())
    self._finished[chain_id].acquire()
    self._samples.append(sample_cache)
    self._treedefs.append(treedef)
    self._samples_count.append(0)
    self._leafnames.append(["/".join(key_tuple) for key_tuple in pytree_keys])
    return chain_id

  def save(self, chain_id: int, values):
    """Saves new leaves to dataset.

    Args:
      chain_id: ID from register_chain
      values: Tree leaves of sample

    """
    sample_cache = self._samples[chain_id]
    for leaf, value in zip(sample_cache, values):
      leaf[self._samples_count[chain_id]] = value
    self._samples_count[chain_id] += 1

  def finalize(self, chain_id: int):
    """Waits for all writing to be finished (scheduled via host_callback).

    Args:
      chain_id: ID of the chain which is finished

    """
    # Is called after all host callback calls have been processed
    self._finished[chain_id].release()
    if self._dir:
      output_dir = Path(self._dir)
      output_file = output_dir / f"chain_{chain_id}.npz"
      output_dir.mkdir(exist_ok=True)
      onp.savez(
        output_file,
        **dict(zip(self._leafnames[chain_id], self._samples[chain_id])))

  def finished(self, chain_id):
    """Returns samples after all data has been processed.

    Finalize is scheduled via host_callback and finished is called in the normal
    python flow. Via a barrier it is possible to pause the program flow until
    all asynchronously saved data has been processed.

    Args:
      chain_id: ID of chain requesting continuation of normal program flow.

    Returns:
      Returns the collected samples in the original tree format but with numpy-
      arrays as leaves.

    """
    self._finished[chain_id].acquire()
    self._finished[chain_id].release()
    # Restore original tree shape
    return tree_util.tree_unflatten(
      self._treedefs[chain_id],
      self._samples[chain_id])

  def checkpoint(self, chain_id: int, state):
    """Called every nth step. """

  def resume(self):
    """Called to restore data loader state and return sample states. """


def load(init_state, checkpoint):
  """Reconstructs an earlier checkpoint."""
  raise NotImplementedError("Checkpointing is currently not supported.")

def save(data_collector: DataCollector = None,
         checkpoint_every: int = 0
         ) -> Saving:
  """Initializes asynchronous saving of samples and checkpoints.

  Accepted samples are sent to the host and processed there. This optimizes the
  memory usage drastically and also allows gaining insight in the data while the
  simulation is running.

  Returns statistics and samples depending on the Data Collector. For
  example hdf5 can be used for samples larger than the (device-)memory.
  Therefore, no samples are returned. Instead, the memory collector returns
  the samples collected as numpy arrays.

  Example usage:

    .. doctest::

      >>> import jax.numpy as jnp
      >>> from jax.lax import scan
      >>> from jax_sgmc import io, scheduler
      >>>
      >>> dc = io.MemoryCollector()
      >>> init_save, save, postprocess_save = io.save(dc)
      >>>
      >>> def update(saving_state, it):
      ...   saving_state = save(saving_state, jnp.mod(it, 2) == 0, {'iteration': it})
      ...   return saving_state
      >>>
      >>> # The information about the number of collected samples must be defined
      >>> # before the run
      >>> static_information = scheduler.static_information(samples_collected=3)
      >>>
      >>> # The saving function must now the form of the sample which should be saved
      >>> saving_state = init_save({'iteration': jnp.array(0)}, {}, static_information)
      >>> final_state, _ = scan(update, saving_state, jnp.arange(5))
      >>>
      >>> saved_samples = postprocess_save(final_state, None)
      >>> print(saved_samples)
      {'sample_count': Array(3, dtype=int32, weak_type=True), 'samples': {'iteration': array([0, 2, 4], dtype=int32)}}


  Args:
    data_collector: Stateful object for data storage and serialization
    checkpoint_every: Create a checkpoint for late resuming every n iterations

  Warning:
    Checkpointing is currently not supported.

  Returns:
    Returns a saving strategy.

  """

  # Todo: Implement checkpointing
  if checkpoint_every != 0:
    raise NotImplementedError("Checkpointing is not supported yet.")

  # Helper functions for host_callback

  def _save(data, *unused_args):
    chain_ids, data = data

    # id_tap sends batched arguments to the host
    if chain_ids.ndim == 0:
      data_collector.save(chain_ids, data)
    else:
      for idx, chain_id in enumerate(chain_ids):
        data_collector.save(chain_id, [leaf[idx] for leaf in data])

  def _save_wrapper(args) -> int:
    # Use the result to count the number of saved samples. The result must be
    # used to avoid losing the call to Jax's optimizations.
    # Only return the leaves, as tree structure is redundant and requires
    # flattening on the host.
    chain_id, data = args
    flat_args = (chain_id, tree_util.tree_leaves(data))
    counter = host_callback.id_tap(_save, flat_args, result=1)
    return counter

  def init(init_sample, init_checkpoint, static_information) -> saving_state:
    """Initializes the saving state.

    Args:
      init_sample: Determining shape and dtype of collected samples
      init_checkpoint: Determining shape and dtype of checkpointed states
      static_information: Information about e. g. the total count of samples
        collected.

    Returns:
      Returns initial state.
    """
    chain_id = data_collector.register_chain(
      init_sample=init_sample,
      init_checkpoint=init_checkpoint,
      static_information=static_information)
    # The count of saved samples is important to ensure that the callback
    # function is not removed by jax's optimization procedures.
    initial_state = saving_state(
      chain_id=chain_id,
      saved_samples=0,
      data=None)
    return initial_state

  @stop_vmap.stop_vmap
  def _save_helper(keep, state, sample):
    return lax.cond(keep,
                    _save_wrapper,
                    lambda *args: 0,
                    (state.chain_id, sample))

  # Todo: Generalize the saving by contracting the scheduler state and the
  #       solver state to a single checkpointing state.
  def save(state: saving_state,
           keep: jnp.bool_,
           sample: Any,
           scheduler_state: scheduler.scheduler_state = None,
           solver_state: Any = None):
    """Calls the data collector on the host via host callback module."""

    # Save sample if samples is not subject to burn in or discarded by thinning
    saved = _save_helper(keep, state, sample)

    # Todo: Implement checkpointing
    # last_checkpoint = lax.cond(time_for_checkpoint,
    #                            _checkpoint_wrapper,
    #                            lambda *args: last_checkpoint,
    #                            (scheduler_state, soler_state)

    new_state = saving_state(
      chain_id=state.chain_id,
      saved_samples=state.saved_samples + saved,
      data=None)
    return new_state, None

  def postprocess(state: saving_state, unused_saved):
    # Call with host callback to ensure that finalized is called after all other
    # id_tap processes were finished.
    host_callback.id_tap(
      lambda id, *unused: data_collector.finalize(id),
      state.chain_id)
    collected_samples = data_collector.finished(state.chain_id)
    return {"sample_count": state.saved_samples,
            "samples": collected_samples}

  return init, save, postprocess

# Todo: Removed the sample collection, consider removing last argument in
#       postprocess function
def no_save() -> Saving:
  """Does not save the data on the host but return it instead.

  If the samples are small, collection on the device is possible. Samples must
  be copied repeatedly.

  Save keep every second element of a 5 element scan

  .. doctest::

    >>> import jax.numpy as jnp
    >>> from jax.lax import scan
    >>> from jax_sgmc import io, scheduler
    >>>
    >>> init_save, save, postprocess_save = io.no_save()
    >>>
    >>> def update(saving_state, it):
    ...   saving_state = save(saving_state, jnp.mod(it, 2) == 0, {'iteration': it})
    ...   return saving_state
    >>>
    >>> # The information about the number of collected samples must be defined
    >>> # before the run
    >>> static_information = scheduler.static_information(samples_collected=3)
    >>>
    >>> # The saving function must now the form of the sample which should be saved
    >>> saving_state = init_save({'iteration': jnp.array(0)}, None, static_information)
    >>> final_state, _ = scan(update, saving_state, jnp.arange(5))
    >>>
    >>> saved_samples = postprocess_save(final_state, None)
    >>> print(saved_samples)
    {'sample_count': Array(3, dtype=int32, weak_type=True), 'samples': {'iteration': Array([0, 2, 4], dtype=int32)}}

  Returns:
    Returns a saving strategy, which keeps the samples entirely in the
    device's memory.

  """
  def init(init_sample: PyTree,
           unused_solver_state: PyTree,
           static_information: PyTree
           ) -> saving_state:
    """Initializes the saving state.

    Args:
      init_sample: Determining shape and dtype of collected samples
      init_checkpoint: Determining shape and dtype of checkpointed states
      static_information: Information about e. g. the total count of samples
        collected.

    Returns:
      Returns initial state.
    """
    def init_zeros(leaf):
      shape = leaf.shape
      new_shape = tuple(onp.append(static_information.samples_collected, shape))
      new_shape = tree_util.tree_map(int, new_shape)
      return jnp.zeros(new_shape, dtype=leaf.dtype)

    init_data = tree_util.tree_map(init_zeros, init_sample)

    # The chain id is unnecessary
    return saving_state(chain_id=0, saved_samples=0, data=init_data)

  def _update_data_leaf(idx, data_leaf, new_slice):
    return data_leaf.at[idx].set(new_slice)

  def _save_sample(args):
    state, sample = args
    new_data = tree_util.tree_map(
      partial(_update_data_leaf, state.saved_samples),
      state.data,
      sample)
    new_state = saving_state(
      chain_id=state.chain_id,
      saved_samples=state.saved_samples + 1,
      data=new_data)
    return new_state

  @stop_vmap.stop_vmap
  def _save_helper(keep, state, sample):
    return lax.cond(keep,
                    _save_sample,
                    lambda args: args[0],
                    (state, sample))

  def save(state: saving_state,
           keep: jnp.bool_,
           sample: Any,
           **unused_kwargs: Any
           ) -> Any:
    """Determines whether a sample should be saved.

    A sample will be saved it is not subject to burn in and not discarded due
    to thinning.

    """
    new_state = _save_helper(keep, state, sample)
    return new_state, None

  def postprocess(state: saving_state, unused_saved):
    """Return the accepted samples. """
    return {"sample_count": state.saved_samples,
            "samples": state.data}

  return init, save, postprocess



# Copyright 2021 Multiscale Modeling of Fluid Materials, TU Munich
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Load numpy arrays in jit-compiled functions.

The numpy data loader is easy to use if the whole dataset fits into RAM and is
already present as numpy-arrays.

"""
from copy import deepcopy

import math
import itertools
from typing import Tuple, Any, Dict, List

import numpy as onp
import jax.numpy as jnp
import jax
from jax import random

from jax_sgmc.data.core import DeviceDataLoader, HostDataLoader, DataLoader
from jax_sgmc.data.core import MiniBatchInformation
from jax_sgmc.data.core import tree_index
from jax_sgmc.util import Array

PyTree = Any

class NumpyBase(DataLoader):

  def __init__(self, on_device: bool = True, copy=True, **reference_data):
    super().__init__()

    observation_counts = []
    self._reference_data = {}
    for name, array in reference_data.items():
      observation_counts.append(len(array))
      # Transform to jax arrays if on device
      if on_device:
        self._reference_data[name] = jnp.array(array, copy=copy)
      else:
        self._reference_data[name] = onp.array(array, copy=copy)

    if len(observation_counts) != 0:
      # Check same number of observations
      if onp.any(onp.array(observation_counts) != observation_counts[0]):
        raise ValueError("All reference_data arrays must have the same length "
                         "in the first dimension.")

      self._observation_count = observation_counts[0]
    else:
      self._observation_count = 0

  @property
  def _format(self):
    """Returns shape and dtype of a single observation. """
    mb_format = {}
    for name, array in self._reference_data.items():
      # Get the format and dtype of the data
      mb_format[name] = jax.ShapeDtypeStruct(
        dtype=self._reference_data[name].dtype,
        shape=tuple(int(s) for s in array.shape[1:]))
    return mb_format

  @property
  def static_information(self):
    """Returns information about total samples count and batch size. """
    information = {
      "observation_count": self._observation_count
    }
    return information


class DeviceNumpyDataLoader(NumpyBase, DeviceDataLoader):
  """Load complete dataset into memory from multiple numpy arrays.

  This data loader supports checkpointing, starting chains from a well-defined
  state and true random access.

  The pipeline can be constructed directly from numpy arrays:

  .. doctest::

    >>> import numpy as onp
    >>> from jax_sgmc.data.numpy_loader import DeviceNumpyDataLoader
    >>>
    >>> x, y = onp.arange(10), onp.zeros((10, 4, 3))
    >>>
    >>> data_loader = DeviceNumpyDataLoader(name_for_x=x, name_for_y=y)
    >>>
    >>> zero_batch = data_loader.initializer_batch(4)
    >>> for key, value in zero_batch.items():
    ...   print(f"{key}: shape={value.shape}, dtype={value.dtype}")
    name_for_x: shape=(4,), dtype=int32
    name_for_y: shape=(4, 4, 3), dtype=float32

  Args:
    reference_data: Each kwarg-pair is an entry in the returned data-dict.
    copy: Whether to copy the reference data (default True) or only create a
      reference.

  """

  def __init__(self, copy=True, **reference_data):
    super().__init__(on_device=True, copy=copy, **reference_data)

  def init_random_data(self, *args, **kwargs) -> PyTree:
    del args
    key = kwargs.get("key", random.PRNGKey(0))
    return key

  # Todo: Provide shuffling and in_epoch_shuffling too
  def get_random_data(self,
                      state,
                      batch_size
                      ) ->Tuple[PyTree, Tuple[PyTree, MiniBatchInformation]]:
    key, split = random.split(state)
    selection_indices = random.randint(
      split, shape=(batch_size,), minval=0, maxval=self._observation_count)

    selected_observations = tree_index(self._reference_data, selection_indices)
    info = MiniBatchInformation(observation_count=self._observation_count,
                                batch_size=batch_size,
                                mask=jnp.ones(batch_size, dtype=jnp.bool_))

    return key, (selected_observations, info)

  def get_full_data(self) -> Dict:
    return self._reference_data


class NumpyDataLoader(NumpyBase, HostDataLoader):
  """Load complete dataset into memory from multiple numpy arrays.

  This data loader supports checkpointing, starting chains from a well-defined
  state and true random access.

  The pipeline can be constructed directly from numpy arrays:

  .. doctest::

    >>> import numpy as onp
    >>> from jax_sgmc.data.numpy_loader import NumpyDataLoader
    >>>
    >>> x, y = onp.arange(10), onp.zeros((10, 4, 3))
    >>>
    >>> data_loader = NumpyDataLoader(name_for_x=x, name_for_y=y)
    >>>
    >>> zero_batch = data_loader.initializer_batch(4)
    >>> for key, value in zero_batch.items():
    ...   print(f"{key}: shape={value.shape}, dtype={value.dtype}")
    name_for_x: shape=(4,), dtype=int32
    name_for_y: shape=(4, 4, 3), dtype=float32

  Args:
    reference_data: Each kwarg-pair is an entry in the returned data-dict.
    copy: Whether to copy the reference data (default True) or only create a
      reference.

  """

  def __init__(self,
               copy=True,
               **reference_data):
    super().__init__(
      on_device=False,
      copy=copy,
      **reference_data)
    self._chains: List = []

  def save_state(self, chain_id: int) -> PyTree:
    """Returns all necessary information to restore the dataloader state.

    Args:
      chain_id: Each chain can be checkpointed independently.

    Returns:
      Returns necessary information to restore the state of the chain via
      :func:`load_state`.

    """
    # Get the state of all random data generators. All other information will be
    # set by initializing the generator on the same way as before

    chain_data = self._chains[chain_id]
    if chain_data['type'] == 'random':
      data = {key: deepcopy(value)
              for key, value in chain_data.items() if key != 'rng'}
      return {'random': (chain_data['rng'].bit_generator.state, data)}
    elif chain_data['type'] == 'ordered':
      return {'ordered': chain_data['idx_offset']}
    else:
      raise ValueError(f"Chain type {chain_data['type']} is unknown.")

  def load_state(self, chain_id: int, data) -> None:
    """Restores dataloader state from previously computed checkpoint.

    Args:
      chain_id: The chain to restore the state.
      data: Data from :func:`save_state` to restore state of the chain.

    """
    # Restore the state by setting the random number generators to the
    # checkpointed state
    type, value = data.popitem()
    if type == 'random':
      rng_state, chain_data = value
      self._chains[chain_id]['rng'].bit_generator.state = rng_state
      for key, value in chain_data.items():
        self._chains[chain_id][key] = value
    elif type == 'ordered':
      self._chains[chain_id]['idx_offset'] = value
    else:
      raise ValueError(f"Chain type {type} is unknown.")

  def register_random_pipeline(self,
                               cache_size: int = 1,
                               mb_size: int = None,
                               in_epochs: bool = False,
                               shuffle: bool = False,
                               **kwargs: Any) -> int:
    """Register a new chain which draws samples randomly.

    Args:
      cache_size: The number of drawn batches.
      mb_size: The number of observations per batch.
      shuffle: Shuffle dataset instead of drawing randomly from the
        observations.
      in_epochs: Samples returned twice per epoch are marked via mask = 0 (only
        if ``shuffle = True``.
      seed: Set the random seed to start the chain at a well-defined state.

    Returns:
      Returns the id of the new chain.

    """
    # The random state of each chain can be defined unambiguously via the
    # PRNGKey
    if mb_size > self._observation_count:
      raise ValueError(f"The batch size cannot be bigger than the observation "
                       f"count. Provided {mb_size} and "
                       f"{self._observation_count}")
    if not shuffle and in_epochs:
      raise ValueError("in_epochs = True can only be used for shuffle = True.")

    chain_id = len(self._chains)

    seed = kwargs.get("seed", chain_id)
    rng = onp.random.default_rng(
      onp.random.SeedSequence(seed).spawn(1)[0])

    # The indices list must have at least the length equal to the number of
    # observations but should also be a multiple of the mb_size to simplify
    # getting new indices.
    new_chain = {'type': 'random',
                 'rng': rng,
                 'idx_offset': None,
                 'in_epochs': in_epochs,
                 'shuffle': shuffle,
                 'remaining_samples': 0,
                 'draws': math.ceil(self._observation_count / mb_size),
                 'random_indices': None,
                 'mb_size': mb_size,
                 'cache_size': cache_size}

    self._chains.append(new_chain)
    return chain_id

  def register_ordered_pipeline(self,
                                cache_size: int = 1,
                                mb_size: int = None,
                                **kwargs
                                ) -> int:
    """Register a chain which assembles batches in an ordered manner.

    Args:
      cache_size: The number of drawn batches.
      mb_size: The number of observations per batch.
      seed: Set the random seed to start the chain at a well-defined state.

    Returns:
      Returns the id of the new chain.

    """
    assert mb_size <= self._observation_count, \
      (f"The batch size cannot be bigger than the observation count. Provided "
       f"{mb_size} and {self._observation_count}")
    chain_id = len(self._chains)

    new_chain = {'type': 'ordered',
                 'rng': None,
                 'idx_offset': 0,
                 'mb_size': mb_size,
                 'cache_size': cache_size}

    self._chains.append(new_chain)
    return chain_id

  def get_batches(self, chain_id: int) -> PyTree:
    """Draws a batch from a chain.

    Args:
      chain_id: ID of the chain, which holds the information about the form of
        the batch and the process of assembling.

    Returns:
      Returns a batch of batches as registered by
      :func:`register_random_pipeline` or :func:`register_ordered_pipeline` with
      `cache_size` batches holding `mb_size` observations.

    """
    # Data slicing is the same for all methods of random and ordered access,
    # only the indices for slicing differ. The method _get_indices find the
    # correct method for the chain.
    selections_idx, selections_mask = self._get_indices(chain_id)

    # Slice the data and transform into device array.
    selected_observations: Dict[str, Array] = {}
    for key, data in self._reference_data.items():
      if data.ndim == 1:
        selection = jnp.array(data[selections_idx,])
      else:
        selection = jnp.array(data[selections_idx,::])
      selected_observations[key] = selection
    return selected_observations, jnp.array(selections_mask, dtype=jnp.bool_)

  def _get_indices(self, chain_id: int):
    chain = self._chains[chain_id]
    if chain['type'] == 'ordered':
      index_fn = self._ordered_indices
    elif chain['in_epochs']:
      index_fn = self._shuffle_in_epochs
    elif chain['shuffle']:
      index_fn = self._shuffle_indices
    else:
      index_fn = self._draw_indices
    indices, masks = list(zip(*map(
      lambda _: index_fn(chain),
      itertools.repeat(chain_id, self._chains[chain_id]['cache_size']))))
    return onp.array(indices), onp.array(masks, dtype=onp.bool_)

  def _ordered_indices(self, chain):
    idcs = onp.arange(chain['mb_size']) + chain['idx_offset']
    # For consistency also return a mask to mark
    #  the samples returned double.
    mask = onp.arange(chain['mb_size']) + chain['idx_offset'] < self._observation_count

    # Start again at the first sample if all samples have been returned
    if chain['idx_offset'] + chain['mb_size'] >= self._observation_count:
      chain['idx_offset'] = 0
    else:
      chain['idx_offset'] += chain['mb_size']
    # Simply return the first samples again if less samples remain than
    # necessary to fill the cache.
    return onp.mod(idcs, self._observation_count), mask

  def _random_indices(self, chain_id: int) -> Tuple[List, Any]:
    """Returns indices and mask to access random data. """
    chain = self._chains[chain_id]
    if chain['in_epochs']:
      return self._shuffle_in_epochs(chain)
    elif chain['shuffle']:
      return self._shuffle_indices(chain)
    else:
      return self._draw_indices(chain)

  def _draw_indices(self, chain):
    # Randomly choose batches
    selections = chain['rng'].choice(
      onp.arange(0, self._observation_count),
      size=chain['mb_size'],
      replace=True)
    mask = onp.ones(chain['mb_size'], dtype=onp.bool_)
    return selections, mask

  def _shuffle_indices(self, chain):
    floor_draws = math.floor(self._observation_count / chain['mb_size'])
    # The partial valid cache must not be changed when updating the indices
    ceil_draws = floor_draws + 2

    if chain['remaining_samples'] < chain['mb_size']:
      # The indices have to be refreshed. Shuffling is equivalent to drawing
      # without replacement.
      new_indices = chain['rng'].choice(
        onp.arange(0, self._observation_count),
        size=self._observation_count,
        replace=False)

      if chain['random_indices'] is None:
        # Special options for first run
        chain['draws'] = 0
        chain['random_indices'] = onp.zeros(
          ceil_draws * chain['mb_size'], dtype=onp.int_)

      # Update only invalid samples (do not overwrite still valid samples)
      update_idxs = onp.mod(
        onp.arange(self._observation_count)
        + chain['draws'] * chain['mb_size']
        + chain['remaining_samples'],
        ceil_draws * chain['mb_size'])
      chain['random_indices'][update_idxs] = new_indices
      chain['remaining_samples'] += self._observation_count

    # All samples are valid
    mask = onp.ones(chain['mb_size'], dtype=onp.bool_)

    # Take the new indices
    selections_idxs = onp.mod(
      onp.arange(chain['mb_size']) + chain['draws'] * chain['mb_size'],
      chain['mb_size'] * ceil_draws)
    selections = onp.copy(chain['random_indices'][selections_idxs])
    chain['draws'] = (chain['draws'] + 1) % ceil_draws
    chain['remaining_samples'] -= chain['mb_size']

    return selections, mask

  def _shuffle_in_epochs(self, chain):
    ceil_draws = math.ceil(self._observation_count / chain['mb_size'])

    if chain['draws'] == ceil_draws:
      # The indices have to be refreshed. Shuffling is equivalent to drawing
      # without replacement.
      new_indices = chain['rng'].choice(
        onp.arange(0, self._observation_count),
        size=self._observation_count,
        replace=False)

      if chain['random_indices'] is None:
        # Special options for first run
        chain['draws'] = 0
        chain['random_indices'] = onp.zeros(
          ceil_draws * chain['mb_size'], dtype=onp.int_)

      chain['random_indices'][0:self._observation_count] = new_indices
      chain['draws'] = 0

    start_idx = chain['mb_size'] * chain['draws']
    end_idx = chain['mb_size'] * (chain['draws'] + 1)

    mask = onp.arange(start_idx, end_idx) < self._observation_count

    selections = onp.copy(chain['random_indices'][start_idx:end_idx])
    chain['draws'] += 1

    return selections, mask



# Copyright 2021 Multiscale Modeling of Fluid Materials, TU Munich
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Load Tensorflow-Datasets in jit-compiled functions.

The tensorflow dataloader supports tensorflow Datasets, e.g. from the
`tensorflow_datasets` package.

Note:
  This submodule requires that ``tensorflow`` and ``tensorflow_datasets`` are
  installed. Additional information can be found in the
  :ref:`installation instructions<additional_requirements>`.

"""

from typing import List, Any

import jax
import jax.numpy as jnp
from jax import tree_util

from tensorflow import data as tfd
import tensorflow_datasets as tfds

from jax_sgmc.data.core import HostDataLoader

PyTree = Any
TFDataSet = tfd.Dataset

class TensorflowDataLoader(HostDataLoader):
  """Load data from a tensorflow dataset object.

  The tensorflow datasets package provides a high number of ready to go
  datasets, which can be provided directly to the Tensorflow Data Loader.

  ::

    import tensorflow_datasets as tdf
    import tensorflow_datasets as tfds
    from jax_sgmc import data
    from jax_sgmc.data.tensorflow_loader import TensorflowDataLoader

    pipeline = tfds.load("cifar10", split="train")
    data_loader = TensorflowDataLoader(pipeline, shuffle_cache=100, exclude_keys=['id'])

  Args:
    pipeline: A tensorflow data pipeline, which can be obtained from the
      tensorflow dataset package

  """

  def __init__(self,
               pipeline: TFDataSet,
               mini_batch_size: int = None,
               shuffle_cache: int = 100,
               exclude_keys: List = None):
    super().__init__()
    # Tensorflow is in general not required to use the library
    assert mini_batch_size is None, "Depreceated"
    assert TFDataSet is not None, "Tensorflow must be installed to use this " \
                                  "feature."
    assert tfds is not None, "Tensorflow datasets must be installed to use " \
                             "this feature."

    self._observation_count = jnp.int32(pipeline.cardinality().numpy())
    # Basic pipeline, from which all other pipelines are constructed
    self._pipeline = pipeline
    self._exclude_keys = [] if exclude_keys is None else exclude_keys
    self._shuffle_cache = shuffle_cache

    self._pipelines: List[TFDataSet] = []

  def register_random_pipeline(self,
                               cache_size: int = 1,
                               mb_size: int = None,
                               **kwargs) -> int:
    """Register a new chain which draws samples randomly.

    Args:
      cache_size: The number of drawn batches.
      mb_size: The number of observations per batch.

    Returns:
      Returns the id of the new chain.

    """

    # Assert that not kwargs are passed with the intention to control the
    # initial state of the tensorflow data loader
    assert kwargs == {}, "Tensorflow data loader does not accept additional "\
                         "kwargs"
    new_chain_id = len(self._pipelines)

    # Randomly draw a number of cache_size mini_batches, where each mini_batch
    # contains self.mini_batch_size elements.
    random_data = self._pipeline.repeat()
    random_data = random_data.shuffle(self._shuffle_cache)
    random_data = random_data.batch(mb_size)
    random_data = random_data.batch(cache_size)

    # The data must be transformed to numpy arrays, as most numpy arrays can
    # be transformed to the duck-typed jax array form
    # random_data = tfds.as_numpy(random_data)
    random_data = iter(random_data)

    self._pipelines.append(random_data)

    return new_chain_id

  def register_ordered_pipeline(self,
                                cache_size: int = 1,
                                mb_size: int = None,
                                **kwargs
                                ) -> int:
    """Register a chain which assembles batches in an ordered manner.

    Args:
      cache_size: The number of drawn batches.
      mb_size: The number of observations per batch.

    Returns:
      Returns the id of the new chain.

    """
    raise NotImplementedError

  def get_batches(self, chain_id: int) -> PyTree:
    """Draws a batch from a chain.

    Args:
      chain_id: ID of the chain, which holds the information about the form of
        the batch and the process of assembling.

    Returns:
      Returns a batch of batches as registered by :func:`register_random_pipeline` or
      :func:`register_ordered_pipeline` with `cache_size` batches holding
      `mb_size` observations.

    """

    # Not supported data types, such as strings, must be excluded before
    # transformation to jax types.
    numpy_batch = next(self._pipelines[chain_id])
    if self._exclude_keys is not None:
      for key in self._exclude_keys:
        del numpy_batch[key]

    return tree_util.tree_map(jnp.array, numpy_batch), None

  @property
  def _format(self):
    data_spec = self._pipeline.element_spec
    if self._exclude_keys is not None:
      not_excluded_elements = {id: elem for id, elem in data_spec.items()
                               if id not in self._exclude_keys}
    else:
      not_excluded_elements = data_spec

    def leaf_dtype_struct(leaf):
      shape = tuple(int(s) for s in leaf.shape if s is not None)
      dtype = leaf.dtype.as_numpy_dtype
      return jax.ShapeDtypeStruct(
        dtype=dtype,
        shape=shape)

    return tree_util.tree_map(leaf_dtype_struct, not_excluded_elements)

  @property
  def static_information(self):
    """Returns information about total samples count and batch size. """
    information = {
      "observation_count" : self._observation_count
    }
    return information



from functools import partial, wraps

import jax
from jax import tree_util
import jax.numpy as jnp

def pytree_list_to_leaves(pytrees):
  """Transform a list of pytrees to allow pmap/vmap.

  The trees must have the same tree structure and only differ in the value of
  their leaves. This means, that the trees might contain custom nodes, such as
  :class:`jax.tree_util.Partial`, but those tree nodes must equivalent. For
  example

  .. doctest::

    >>> from jax.tree_util import Partial
    >>>
    >>> Partial(lambda x: x + 1) == Partial(lambda x: x + 1)
    False

  because they are defined on different functions, but are still equivalent as
  the functions perform the same computations.

  Example usage:

  .. doctest::

    >>> import jax.numpy as jnp
    >>> import jax_sgmc.util.list_map as lm
    >>>
    >>> tree_a = {"a": 0.0, "b": jnp.zeros((2,))}
    >>> tree_b = {"a": 1.0, "b": jnp.ones((2,))}
    >>>
    >>> concat_tree = lm.pytree_list_to_leaves([tree_a, tree_b])
    >>> print(concat_tree)
    {'a': Array([0., 1.], dtype=float32, weak_type=True), 'b': Array([[0., 0.],
           [1., 1.]], dtype=float32)}


  Args:
    pytrees: A list of trees with similar tree structure and equally shaped
      leaves

  Returns:
    Returns a tree with the same tree structure but corresponding leaves
    concatenated along the first dimension.

  """

  # Transpose the pytrees, i. e. make a list (array) of leaves from a list of
  # pytrees. Only then vmap can be used to vectorize an operation over pytrees
  treedef = tree_util.tree_structure(pytrees[0])
  superleaves = [jnp.stack(leaves, axis=0)
                 for leaves in zip(*map(tree_util.tree_leaves, pytrees))]
  return tree_util.tree_unflatten(treedef, superleaves)


def pytree_leaves_to_list(pytree):
  """Splits a pytree in a list of pytrees.

  Splits every leaf of the pytree along the first dimension, thus undoing the
  :func:`pytree_list_to_leaves` transformation.

  Example usage:

  .. doctest::

    >>> import jax.numpy as jnp
    >>> import jax_sgmc.util.list_map as lm
    >>>
    >>> tree = {"a": jnp.array([0.0, 1.0]), "b": jnp.zeros((2, 2))}
    >>>
    >>> tree_list = lm.pytree_leaves_to_list(tree)
    >>> print(tree_list)
    [{'a': Array(0., dtype=float32), 'b': Array([0., 0.], dtype=float32)}, {'a': Array(1., dtype=float32), 'b': Array([0., 0.], dtype=float32)}]


  Args:
    pytree: A single pytree where each leaf has eqal `leaf.shape[0]`.

  Returns:
    Returns a list of pytrees with similar structure.

  """
  leaves, treedef = tree_util.tree_flatten(pytree)
  num_trees = leaves[0].shape[0]
  pytrees = [tree_util.tree_unflatten(treedef, [leaf[idx] for leaf in leaves])
             for idx in range(num_trees)]
  return pytrees


def list_vmap(fun):
  """vmaps a function over similar pytrees.

  Example usage:

  .. doctest::

    >>> from jax import tree_map
    >>> import jax.numpy as jnp
    >>> import jax_sgmc.util.list_map as lm
    >>>
    >>> tree_a = {"a": 0.0, "b": jnp.zeros((2,))}
    >>> tree_b = {"a": 1.0, "b": jnp.ones((2,))}
    >>>
    ... @lm.list_vmap
    ... def tree_add(pytree):
    ...   return tree_map(jnp.subtract, pytree, tree_b)
    >>>
    >>> print(tree_add(tree_a, tree_b))
    [{'a': Array(-1., dtype=float32, weak_type=True), 'b': Array([-1., -1.], dtype=float32)}, {'a': Array(0., dtype=float32, weak_type=True), 'b': Array([0., 0.], dtype=float32)}]

  Args:
    fun: Function accepting a single pytree as first argument.

  Returns:
    Returns a vmapped-function accepting multiple pytree args with similar tree-
    structure.

  """
  vmap_fun = jax.vmap(fun, 0, 0)
  @wraps(fun)
  def vmapped(*pytrees):
    single_tree = pytree_list_to_leaves(pytrees)
    single_result = vmap_fun(single_tree)
    return pytree_leaves_to_list(single_result)
  return vmapped

def list_pmap(fun):
  """pmaps a function over similar pytrees.

  Args:
    fun: Function accepting a single pytree as first argument.

  Returns:
    Returns a pmapped-function accepting multiple pytree args with similar tree-
    structure.

  """
  pmap_fun = jax.pmap(fun, 0)
  @wraps(fun)
  def pmapped(*pytrees):
    single_tree = pytree_list_to_leaves(pytrees)
    single_result = pmap_fun(single_tree)
    return pytree_leaves_to_list(single_result)
  return pmapped



# Copyright 2021 Multiscale Modeling of Fluid Materials, TU Munich
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines types special to jax.edited.bak or this library. """

from typing import Any, NamedTuple
from functools import partial

from jax import tree_util
from jax import flatten_util
import jax.numpy as jnp

Array = jnp.ndarray
PyTree = Any

class Tensor(NamedTuple):
  """Vector and matrix pytree-products.

  Attributes:
    ndim: Dimension of the pytree (1: vector, 2: matrix)
    tensor: Data of the pytree

  """
  ndim: int
  tensor: PyTree

def tensor_matmul(matrix: Tensor, vector: PyTree):
  """Matrix vector product with a tensor and a pytree.

  Distinguishes between full matrices and diagonal matrices.

  Args:
    matrix: Matrix in tensor format
    vector: PyTree, which is compatible to the tensor

  """
  if matrix.ndim == 0:
    return tree_scale(matrix.tensor, vector)
  elif matrix.ndim == 1:
    return tree_multiply(matrix.tensor, vector)
  elif matrix.ndim == 2:
    return tree_matmul(matrix.tensor, vector)
  else:
    raise NotImplementedError(f"Cannot multiply matrix with dimension "
                              f"{matrix.ndim}")

def tree_multiply(tree_a: PyTree, tree_b: PyTree) -> PyTree:
  """Maps elementwise product over two vectors.

  Args:
    a: First pytree
    b: Second pytree, must have the same shape as a

  Returns:
    Returns a PyTree obtained by an element-wise product of all PyTree leaves.

  """
  return tree_util.tree_map(jnp.multiply, tree_a, tree_b)


def tree_scale(alpha: Array, tree: PyTree) -> PyTree:
  """Scalar-Pytree product via tree_map.

  Args:
    alpha: Scalar
    a: Arbitrary PyTree

  Returns:
    Returns a PyTree with all leaves scaled by alpha.

  """
  @partial(partial, tree_util.tree_map)
  def tree_scale_imp(x: PyTree):
    return alpha * x
  return tree_scale_imp(tree)


def tree_add(tree_a: PyTree, tree_b: PyTree) -> PyTree:
  """Maps elementwise sum over PyTrees.

  Arguments:
    a: First PyTree
    b: Second PyTree with the same shape as a

  Returns:
    Returns a PyTree obtained by leave-wise summation.
  """
  @partial(partial, tree_util.tree_map)
  def tree_add_imp(leaf_a, leaf_b):
    return leaf_a + leaf_b
  return tree_add_imp(tree_a, tree_b)


def tree_matmul(tree_mat: Array, tree_vec: PyTree):
  """Matrix tree product for LD on manifold.

  Arguments:
    tree_mat: Matrix to be multiplied with flattened tree
    tree_vec: Tree representing vector

  Returns:
    Returns the un-flattened product of the matrix and the flattened tree.
  """
  # Todo: Redefine without need for flatten util
  vec_flat, unravel_fn = flatten_util.ravel_pytree(tree_vec)
  return unravel_fn(jnp.matmul(tree_mat, vec_flat))

def tree_dot(tree_a: PyTree, tree_b: PyTree):
  """Scalar product of two pytrees.

  Args:
    tree_a: First pytree
    tree_b: Second pytree with same tree stree structure and leaf shape as
      tree_a

  Returns:
    Returns a scalar, which is the sum of the element-wise product of all leaves.

  """
  leaves_a = tree_util.tree_leaves(tree_a)
  leaves_b = tree_util.tree_leaves(tree_b)
  return sum((jnp.sum(jnp.multiply(a, b)) for a, b in zip(leaves_a, leaves_b)))
